"""sample html for pytesting"""
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

sample_html = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN""http://www.w3.org/TR/html4/loose.dtd">
<html>
<head>
<title> 1050PageMillRoad Mountain View Moffett Fld Nas CA USA TMY3 WMO#=745090
  2014-01-13
  16:47:19 
 - EnergyPlus</title>
</head>
<body>
<p><a href="#toc" style="float: right">Table of Contents</a></p>
<a name=top></a>
<p>Program Version:<b>EnergyPlus-Windows-32 8.1.0.008, YMD=2014.01.13 16:47</b></p>
<p>Tabular Output Report in Format: <b>HTML</b></p>
<p>Building: <b>1050PageMillRoad</b></p>
<p>Environment: <b>Mountain View Moffett Fld Nas CA USA TMY3 WMO#=745090</b></p>
<p>Simulation Timestamp: <b>2014-01-13
  16:47:19</b></p>
<hr>
<p><a href="#toc" style="float: right">Table of Contents</a></p>
<a name=AnnualBuildingUtilityPerformanceSummary::EntireFacility></a>
<p>Report:<b> Annual Building Utility Performance Summary</b></p>
<p>For:<b> Entire Facility</b></p>
<p>Timestamp: <b>2014-01-13
    16:47:19</b></p>
<b>Values gathered over      8760.00 hours</b><br><br>
<b></b><br><br>
<b>Site and Source Energy</b><br><br>
<!-- FullName:Annual Building Utility Performance Summary_Entire Facility_Site and Source Energy-->
<table border="1" cellspacing="0" cellpadding="4">
    <tr>
        <td>a</td>
        <td>2</td>
    </tr>
    <tr>
        <td>3</td>
        <td>4</td>
    </tr>
</table><br><br>
<b>Site to Source Energy Conversion Factors</b><br><br>
<!-- FullName:Annual Building Utility Performance Summary_Entire Facility_Site to Source Energy Conversion Factors-->
<table border="1" cellspacing="0" cellpadding="4">
    <tr>
        <td>b</td>
        <td>6</td>
    </tr>
    <tr>
        <td>7</td>
        <td>8</td>
    </tr>
</table><br><br>
<p>Some:<b> Random stuff</b></p>

<hr>
<p>Report:<b> COMPONENTS OF PEAK ELECTRICAL DEMAND</b></p>
<p>For:<b> Meter</b></p>
<p>Timestamp: <b>2014-01-13
    16:47:19</b></p>
<b>Custom Monthly Report</b><br><br>
<table border="1" cellspacing="0" cellpadding="4">
    <tr>
        <td>c</td>
        <td>16</td>
    </tr>
    <tr>
        <td>17</td>
        <td>18</td>
    </tr>
</table><br><br>

<hr>
<p>Report:<b> COMPONENTS OF PEAK NET ELECTRICAL DEMAND</b></p>
<p>For:<b> Meter</b></p>
<p>Timestamp: <b>2014-01-13
    16:47:19</b></p>
<b>Custom Monthly Report</b><br><br>
<table border="1" cellspacing="0" cellpadding="4">
    <tr>
        <td>d</td>
        <td>26</td>
    </tr>
    <tr>
        <td>27</td>
        <td>28</td>
    </tr>
</table><br><br>


<br><br>
<b>Computation -  Automatic</b><br><br>
MONTHLYRATEGASCHARGE FROM TotalEnergy CA_MONTHLYGASRATES<br>
EnergyCharges SUM MONTHLYRATEGASCHARGE<br>
Basis SUM EnergyCharges DemandCharges ServiceCharges<br>
Subtotal SUM Basis Adjustment Surcharge<br>
TAXOFEIGHTPERCENT FROM Subtotal<br>
Taxes SUM TAXOFEIGHTPERCENT<br>
Total SUM Subtotal Taxes<br>
</body>
</html>
"""



# Copyright (c) 2012 Santosh Philip
# Copyright (c) 2020 Cheng Cui
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""helper functions for the functions called by bunchdt"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import itertools
from itertools import zip_longest
from eppy.constructions import thermal_properties
from eppy.geometry import surface as g_surface
import eppy.fanpower


def grouper(num, iterable, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * num
    return zip_longest(fillvalue=fillvalue, *args)


def getcoords(ddtt):
    """return the coordinates of the surface"""
    n_vertices_index = ddtt.objls.index("Number_of_Vertices")
    first_x = n_vertices_index + 1  # X of first coordinate
    pts = ddtt.obj[first_x:]
    return list(grouper(3, pts))


def area(ddtt):
    """area of the surface"""
    coords = getcoords(ddtt)
    return g_surface.area(coords)


def height(ddtt):
    """height of the surface"""
    coords = getcoords(ddtt)
    return g_surface.height(coords)


def width(ddtt):
    """width of the surface"""
    coords = getcoords(ddtt)
    return g_surface.width(coords)


def azimuth(ddtt):
    """azimuth of the surface"""
    coords = getcoords(ddtt)
    return g_surface.azimuth(coords)


def true_azimuth(ddtt):
    """true azimuth of the surface"""
    idf = ddtt.theidf
    coord_system = idf.idfobjects["GlobalGeometryRules"][0].Coordinate_System
    if coord_system.lower() == "relative":
        zone_name = ddtt.Zone_Name
        bldg_north = idf.idfobjects["Building"][0].North_Axis
        zone_rel_north = idf.getobject("Zone", zone_name).Direction_of_Relative_North
        surf_azimuth = azimuth(ddtt)
        return g_surface.true_azimuth(bldg_north, zone_rel_north, surf_azimuth)
    elif coord_system.lower() in ("world", "absolute"):
        # NOTE: "absolute" is not supported in v9.3.0
        return azimuth(ddtt)
    else:
        raise ValueError(
            "'{:s}' is no valid value for 'Coordinate System'".format(coord_system)
        )


def tilt(ddtt):
    """tilt of the surface"""
    coords = getcoords(ddtt)
    return g_surface.tilt(coords)


def buildingname(ddtt):
    """return building name"""
    idf = ddtt.theidf
    building = idf.idfobjects["building".upper()][0]
    return building.Name


def zonesurfaces(ddtt):
    """return al list of surfaces that belong to the zone"""
    kwargs = {"fields": ["Zone_Name"], "iddgroups": ["Thermal Zones and Surfaces"]}
    return ddtt.getreferingobjs(**kwargs)


def subsurfaces(ddtt):
    """return al list of surfaces that belong to the zone"""
    kwargs = {
        "fields": ["Building_Surface_Name"],
        "iddgroups": ["Thermal Zones and Surfaces"],
    }
    return ddtt.getreferingobjs(**kwargs)


def rvalue(ddtt):
    return thermal_properties.rvalue(ddtt)


def ufactor(ddtt):
    return thermal_properties.ufactor(ddtt)


def ufactor_ip(ddtt):
    return thermal_properties.ufactor_ip(ddtt)


def rvalue_ip(ddtt):
    return thermal_properties.rvalue_ip(ddtt)


def heatcapacity(ddtt):
    return thermal_properties.heatcapacity(ddtt)


def fanpower_bhp(ddtt):
    """return fanpower in bhp"""
    return eppy.fanpower.fanpower_bhp(ddtt)


def fanpower_watts(ddtt):
    """return fanpower in watts"""
    return eppy.fanpower.fanpower_watts(ddtt)


def fan_maxcfm(ddtt):
    """return the Maximum_Flow_Rate in cfm"""
    return eppy.fanpower.fan_maxcfm(ddtt)



# Copyright (c) 2020 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
# -*- coding: utf-8 -*-
"""functions to do a fast read from the E+ HTML table file"""
# TODO : move this to eppy.readhtml
# TODO Document it in user documentation.
import copy
from io import StringIO

from eppy.results import readhtml


def _decodeline(line, encoding="utf-8"):
    """decodes bytes to string, if line is not bytes, line is returned

    It will first attempt to decode line with value of `encoding`. If that fails, it will try with encoding="ISO-8859-2". If that fails, it will return line.

    Why is it trying encoding="ISO-8859-2". Looks like E+ uses this encoding in some example files and which is then output in the HTML file

    # TODO this code looks fragile. Maybe use standard library HTML parse to deal with encoding?

    Parameters
    ----------
    line : str, bytes
    encoding : str

    Returns
    -------
    line : str
        decoded line
    """
    try:
        return line.decode(encoding)
    except (AttributeError, UnicodeDecodeError) as e:
        if e.__class__ == UnicodeDecodeError:
            # encoding could be ISO-8859-2 in e+ html
            return _decodeline(line, encoding="ISO-8859-2")
        else:
            return line


def getnexttable(fhandle):
    """get the next table in the html file

    Continues to read the file line by line and collects lines from the start of the next table until the end of the table

    Parameters
    ----------
    fhandle : file like object
        A file handle to the E+ HTML table file

    Returns
    -------
    table : str
        The table in HTML format
    """
    lines = fhandle
    tablelines = []
    for line in lines:
        line = _decodeline(line)
        if line.strip().startswith("<table"):
            tablelines.append(line)
            break
    for line in lines:
        line = _decodeline(line)
        tablelines.append(line)
        if line.strip().startswith("</table"):
            break
    return "".join(tablelines)


def tablebyname(filehandle, header):
    """fast extraction of the table using the header to identify the table

    This function reads only one table from the HTML file. This is in contrast to `results.readhtml.titletable` that will read all the tables into memory and allows you to interactively look thru them. The function `results.readhtml.titletable` can be very slow on large HTML files.

    This function is useful when you know which file you are looking for. It looks for the title line that is in bold just before the table. Some tables don't have such a title in bold. This function will not work for tables that don't have a title in bold

    Parameters
    ----------
    fhandle : file like object
        A file handle to the E+ HTML table file
    header: str
        This is the title of the table you are looking for

    Returns
    -------
    titleandtable : (str, list)
        - (title, table)
            - title = previous item with a <b> tag
            - table = rows -> [[cell1, cell2, ..], [cell1, cell2, ..], ..]
    """
    htmlheader = f"<b>{header}</b><br><br>"

    with filehandle:
        for line in filehandle:
            line = _decodeline(line)
            if line.strip() == htmlheader:
                justtable = getnexttable(filehandle)
                thetable = f"{htmlheader}\n{justtable}"
                break

    filehandle = StringIO(thetable)
    htables = readhtml.titletable(filehandle)
    try:
        return list(htables[0])
    except IndexError as e:
        None


def get_upto_nexttable(fhandle):
    """get all lines from the present location in fhandle to the end of the next table

    This function is used by `tablebyindex` to find the title for the table, which is in the lines before the table. Then it can return the title and the table

    Parameters
    ----------
    fhandle : file like object
        A file handle to the E+ HTML table file

    Returns
    -------
    lines_and_table : str
        The table in HTML format with lines before it.
    """
    lines = fhandle
    tablelines = []
    for line in lines:
        line = _decodeline(line)
        tablelines.append(line)
        if line.strip().startswith("</table"):
            break
    return "".join(tablelines)


def tablebyindex(filehandle, index):
    """fast extraction of the table using the index to identify the table

    This function reads only one table from the HTML file. This is in contrast to `results.readhtml.titletable` that will read all the tables into memory and allows you to interactively look thru them. The function `results.readhtml.titletable` can be very slow on large HTML files.

    This function is useful when you know which file you are looking for. It does not work with negative indices, like you can in a list. If you know a way to make negative indices work, do a pull request :-)

    Parameters
    ----------
    fhandle : file like object
        A file handle to the E+ HTML table file
    index: int
        This is the index of the table you are looking for

    Returns
    -------
    titleandtable : (str, list)
        - (title, table)
            - title = previous item with a <b> tag
            - table = rows -> [[cell1, cell2, ..], [cell1, cell2, ..], ..]
    """
    with filehandle:
        tableindex = 0
        for i in range(index + 1):
            thetable = get_upto_nexttable(filehandle)
    filehandle = StringIO(thetable)
    htables = readhtml.titletable(filehandle)
    try:
        return htables[0]
    except IndexError as e:
        None



#!/usr/bin/env python

## EPlusInterface (EPI) - An interface for EnergyPlus
## Copyright (C) 2004 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""legacy code from EPlusInterface"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO
from io import FileIO
from decorator import decorator

import eppy.EPlusInterfaceFunctions.mylib1 as mylib1
import eppy.EPlusInterfaceFunctions.mylib2 as mylib2
import eppy.EPlusInterfaceFunctions.iddgroups as iddgroups
import eppy.EPlusInterfaceFunctions.iddindex as iddindex


def nocomment(astr, com):
    """
    just like the comment in python.
    removes any text after the phrase 'com'
    """
    alist = astr.splitlines()
    for i in range(len(alist)):
        element = alist[i]
        pnt = element.find(com)
        if pnt != -1:
            alist[i] = element[:pnt]
    return "\n".join(alist)


def get_nocom_vars(astr):
    """
    input 'astr' which is the Energy+.idd file as a string
    returns (st1, st2, lss)
    st1 = with all the ! comments striped
    st2 = strips all comments - both the '!' and '\\'
    lss = nested list of all the variables in Energy+.idd file
    """
    nocom = nocomment(astr, "!")  # remove '!' comments
    st1 = nocom
    nocom1 = nocomment(st1, "\\")  # remove '\' comments
    st1 = nocom
    st2 = nocom1
    # alist = string.split(st2, ';')
    alist = st2.split(";")
    lss = []

    # break the .idd file into a nested list
    # =======================================
    for element in alist:
        # item = string.split(element, ',')
        item = element.split(",")
        lss.append(item)
    for i in range(0, len(lss)):
        for j in range(0, len(lss[i])):
            lss[i][j] = lss[i][j].strip()
    if len(lss) > 1:
        lss.pop(-1)
    # =======================================

    # st1 has the '\' comments --- looks like I don't use this
    # lss is the .idd file as a nested list
    return (st1, st2, lss)


def removeblanklines(astr):
    """remove the blank lines in astr"""
    lines = astr.splitlines()
    lines = [line for line in lines if line.strip() != ""]
    return "\n".join(lines)


def _readfname(fname):
    """copied from extractidddata below.
    It deals with all the types of fnames"""
    try:
        if isinstance(fname, (file, StringIO)):
            astr = fname.read()
        else:
            astr = open(fname, "rb").read()
    except NameError:
        if isinstance(fname, (FileIO, StringIO)):
            astr = fname.read()
        else:
            astr = mylib2.readfile(fname)
    return astr


@decorator
def make_idd_index(extract_func, fname, debug):
    """generate the iddindex"""
    astr = _readfname(fname)

    # fname is exhausted by the above read
    # reconstitute fname as a StringIO
    fname = StringIO(astr)

    # glist = iddgroups.iddtxt2grouplist(astr.decode('ISO-8859-2'))

    blocklst, commlst, commdct = extract_func(fname)

    name2refs = iddindex.makename2refdct(commdct)
    ref2namesdct = iddindex.makeref2namesdct(name2refs)
    idd_index = dict(name2refs=name2refs, ref2names=ref2namesdct)
    commdct = iddindex.ref2names2commdct(ref2namesdct, commdct)

    return blocklst, commlst, commdct, idd_index


@decorator
def embedgroupdata(extract_func, fname, debug):
    """insert group info into extracted idd"""

    astr = _readfname(fname)

    # fname is exhausted by the above read
    # reconstitute fname as a StringIO
    fname = StringIO(astr)

    try:
        astr = astr.decode("ISO-8859-2")
    except Exception as e:
        pass  # for python 3
    glist = iddgroups.iddtxt2grouplist(astr)

    blocklst, commlst, commdct = extract_func(fname)
    # add group information to commlst and commdct
    # glist = getglist(fname)
    commlst = iddgroups.group2commlst(commlst, glist)
    commdct = iddgroups.group2commdct(commdct, glist)
    return blocklst, commlst, commdct


@make_idd_index
@embedgroupdata
def extractidddata(fname, debug=False):
    """
    extracts all the needed information out of the idd file
    if debug is True,  it generates a series of text files.
    Each text file is incrementally different. You can do a diff
    see what the change is
    -
    this code is from 2004.
    it works.
    I am trying not to change it (until I rewrite the whole thing)
    to add functionality to it, I am using decorators
    So if
    Does not integrate group data into the results (@embedgroupdata does it)
    Does not integrate iddindex into the results (@make_idd_index does it)
    """
    try:
        if isinstance(fname, (file, StringIO)):
            astr = fname.read()
            try:
                astr = astr.decode("ISO-8859-2")
            except AttributeError:
                pass
        else:
            astr = mylib2.readfile(fname)
            # astr = astr.decode('ISO-8859-2') -> mylib1 does a decode
    except NameError:
        if isinstance(fname, (FileIO, StringIO)):
            astr = fname.read()
            try:
                astr = astr.decode("ISO-8859-2")
            except AttributeError:
                pass
        else:
            astr = mylib2.readfile(fname)
            # astr = astr.decode('ISO-8859-2') -> mylib2.readfile has decoded
    (nocom, nocom1, blocklst) = get_nocom_vars(astr)

    astr = nocom
    st1 = removeblanklines(astr)
    if debug:
        mylib1.write_str2file("nocom2.txt", st1.encode("latin-1"))

    # find the groups and the start object of the group
    # find all the group strings
    groupls = []
    alist = st1.splitlines()
    for element in alist:
        lss = element.split()
        if lss[0].upper() == "\\group".upper():
            groupls.append(element)

    # find the var just after each item in groupls
    groupstart = []
    for i in range(len(groupls)):
        iindex = alist.index(groupls[i])
        groupstart.append([alist[iindex], alist[iindex + 1]])

    # remove the group commentline
    for element in groupls:
        alist.remove(element)

    if debug:
        st1 = "\n".join(alist)
        mylib1.write_str2file("nocom3.txt", st1.encode("latin-1"))

    # strip each line
    for i in range(len(alist)):
        alist[i] = alist[i].strip()

    if debug:
        st1 = "\n".join(alist)
        mylib1.write_str2file("nocom4.txt", st1.encode("latin-1"))

    # ensure that each line is a comment or variable
    # find lines that don't start with a comment
    # if this line has a comment in it
    #   then move the comment to a new line below
    lss = []
    for i in range(len(alist)):
        # find lines that don't start with a comment
        if alist[i][0] != "\\":
            # if this line has a comment in it
            pnt = alist[i].find("\\")
            if pnt != -1:
                # then move the comment to a new line below
                lss.append(alist[i][:pnt].strip())
                lss.append(alist[i][pnt:].strip())
            else:
                lss.append(alist[i])
        else:
            lss.append(alist[i])

    alist = lss[:]
    if debug:
        st1 = "\n".join(alist)
        mylib1.write_str2file("nocom5.txt", st1.encode("latin-1"))

    # need to make sure that each line has only one variable - as in WindowGlassSpectralData,
    lss = []
    for element in alist:
        # if the line is not a comment
        if element[0] != "\\":
            # test for more than one var
            llist = element.split(",")
            if llist[-1] == "":
                tmp = llist.pop()
            for elm in llist:
                if elm[-1] == ";":
                    lss.append(elm.strip())
                else:
                    lss.append((elm + ",").strip())
        else:
            lss.append(element)

    ls_debug = alist[:]  # needed for the next debug - 'nocom7.txt'
    alist = lss[:]
    if debug:
        st1 = "\n".join(alist)
        mylib1.write_str2file("nocom6.txt", st1.encode("latin-1"))

    if debug:
        # need to make sure that each line has only one variable - as in WindowGlassSpectralData,
        # this is same as above.
        # but the variables are put in without the ';' and ','
        # so we can do a diff between 'nocom7.txt' and 'nocom8.txt'. Should be identical
        lss_debug = []
        for element in ls_debug:
            # if the line is not a comment
            if element[0] != "\\":
                # test for more than one var
                llist = element.split(",")
                if llist[-1] == "":
                    tmp = llist.pop()
                for elm in llist:
                    if elm[-1] == ";":
                        lss_debug.append(elm[:-1].strip())
                    else:
                        lss_debug.append((elm).strip())
            else:
                lss_debug.append(element)

        ls_debug = lss_debug[:]
        st1 = "\n".join(ls_debug)
        mylib1.write_str2file("nocom7.txt", st1.encode("latin-1"))

    # replace each var with '=====var======'
    # join into a string,
    # split using '=====var====='
    for i in range(len(lss)):
        # if the line is not a comment
        if lss[i][0] != "\\":
            lss[i] = "=====var====="

    st2 = "\n".join(lss)
    lss = st2.split("=====var=====\n")
    lss.pop(0)  # the above split generates an extra item at start

    if debug:
        fname = "nocom8.txt"
        fhandle = open(fname, "wb")
        k = 0
        for i in range(len(blocklst)):
            for j in range(len(blocklst[i])):
                atxt = blocklst[i][j] + "\n"
                fhandle.write(atxt)
                atxt = lss[k]
                fhandle.write(atxt.encode("latin-1"))
                k = k + 1

        fhandle.close()

    # map the structure of the comments -(this is 'lss' now) to
    # the structure of blocklst - blocklst is a nested list
    # make lss a similar nested list
    k = 0
    lst = []
    for i in range(len(blocklst)):
        lst.append([])
        for j in range(len(blocklst[i])):
            lst[i].append(lss[k])
            k = k + 1

    if debug:
        fname = "nocom9.txt"
        fhandle = open(fname, "wb")
        k = 0
        for i in range(len(blocklst)):
            for j in range(len(blocklst[i])):
                atxt = blocklst[i][j] + "\n"
                fhandle.write(atxt)
                fhandle.write(lst[i][j].encode("latin-1"))
                k = k + 1

        fhandle.close()

    # break up multiple line comment so that it is a list
    for i in range(len(lst)):
        for j in range(len(lst[i])):
            lst[i][j] = lst[i][j].splitlines()
            # remove the '\'
            for k in range(len(lst[i][j])):
                lst[i][j][k] = lst[i][j][k][1:]
    commlst = lst

    # copied with minor modifications from readidd2_2.py -- which has been erased ha !
    clist = lst
    lss = []
    for i in range(0, len(clist)):
        alist = []
        for j in range(0, len(clist[i])):
            itt = clist[i][j]
            ddtt = {}
            for element in itt:
                if len(element.split()) == 0:
                    break
                ddtt[element.split()[0].lower()] = []

            for element in itt:
                if len(element.split()) == 0:
                    break
                # ddtt[element.split()[0].lower()].append(string.join(element.split()[1:]))
                ddtt[element.split()[0].lower()].append(" ".join(element.split()[1:]))

            alist.append(ddtt)

        lss.append(alist)
    commdct = lss

    # add group information to commlst and commdct
    # glist = iddgroups.idd2grouplist(fname)
    # commlst = group2commlst(commlst, glist)
    # commdct = group2commdct(commdct, glist)

    return blocklst, commlst, commdct
    # give blocklst a better name :-(


def getobjectref(blocklst, commdct):
    """
    makes a dictionary of object-lists
    each item in the dictionary points to a list of tuples
    the tuple is (objectname,  fieldindex)
    """
    objlst_dct = {}
    for eli in commdct:
        for elj in eli:
            if "object-list" in elj:
                objlist = elj["object-list"][0]
                objlst_dct[objlist] = []

    for objlist in list(objlst_dct.keys()):
        for i in range(len(commdct)):
            for j in range(len(commdct[i])):
                if "reference" in commdct[i][j]:
                    for ref in commdct[i][j]["reference"]:
                        if ref == objlist:
                            objlst_dct[objlist].append((blocklst[i][0], j))
    return objlst_dct



# Copyright (c) 2012 Tuan Tran
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""pytest for volume_zone.py"""
# Wrote by Tuan Tran trantuan@hawaii.edu / tranhuuanhtuan@gmail.com
# School of Architecture, University of Hawaii at Manoa
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import eppy.geometry.volume_zone as volume_zone
from eppy.pytest_helpers import almostequal


def test_volume():
    """py.test for volume"""
    data = (
        (
            [(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)],
            [(0, 0, 1), (1, 0, 1), (1, 1, 1), (0, 1, 1)],
            1,
        ),  # poly1, poly2, answer
        (
            [(0, 0, 0), (1, 0, 0), (1, 1, 0), (0, 1, 0)],
            [(0, 0, 0), (1, 0, 0), (1, 1, 2), (0, 1, 2)],
            1,
        ),  # poly1, poly2, answer
    )
    for poly1, poly2, answer in data:
        result = volume_zone.vol(poly1, poly2)
        assert almostequal(answer, result, places=4) == True



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

e = [
    (("p_loop Supply Inlet", "epnode"), "Central_Chiller"),
    ("Central_Chiller", ("Central_Chiller_np1_node", "epnode")),
    (("Central_Chiller_np1_node", "epnode"), "np1"),
    ("np1", ("np1_np2_node", "epnode")),
    (("np1_np2_node", "epnode"), "np2"),
    ("np2", ("np2_Outlet_Node_Name", "epnode")),
    (("sb1_pipe_inlet", "epnode"), "sb1_pipe"),
    ("sb1_pipe", ("sb1_pipe_outlet", "epnode")),
    (("sb2_pipe_inlet", "epnode"), "sb2_pipe"),
    ("sb2_pipe", ("sb2_pipe_outlet", "epnode")),
    (("sb3_pipe_inlet", "epnode"), "sb3_pipe"),
    ("sb3_pipe", ("sb3_pipe_outlet", "epnode")),
    (("sb4_pipe_inlet", "epnode"), "sb4_pipe"),
    ("sb4_pipe", ("p_loop Supply Outlet", "epnode")),
    (("p_loop Demand Inlet", "epnode"), "db0_pipe"),
    ("db0_pipe", ("db0_pipe_outlet", "epnode")),
    (("db1_pipe_inlet", "epnode"), "db1_pipe"),
    ("db1_pipe", ("db1_pipe_outlet", "epnode")),
    (("db2_pipe_inlet", "epnode"), "db2_pipe"),
    ("db2_pipe", ("db2_pipe_outlet", "epnode")),
    (("db3_pipe_inlet", "epnode"), "db3_pipe"),
    ("db3_pipe", ("db3_pipe_outlet", "epnode")),
    (("db4_pipe_inlet", "epnode"), "db4_pipe"),
    ("db4_pipe", ("p_loop Demand Outlet", "epnode")),
    (("np2_Outlet_Node_Name", "epnode"), "p_loop_supply_splitter"),
    ("p_loop_supply_splitter", ("sb1_pipe_inlet", "epnode")),
    ("p_loop_supply_splitter", ("sb2_pipe_inlet", "epnode")),
    ("p_loop_supply_splitter", ("sb3_pipe_inlet", "epnode")),
    (("db0_pipe_outlet", "epnode"), "p_loop_demand_splitter"),
    ("p_loop_demand_splitter", ("db1_pipe_inlet", "epnode")),
    ("p_loop_demand_splitter", ("db2_pipe_inlet", "epnode")),
    ("p_loop_demand_splitter", ("db3_pipe_inlet", "epnode")),
    ("p_loop_supply_mixer", ("sb4_pipe_inlet", "epnode")),
    (("sb1_pipe_outlet", "epnode"), "p_loop_supply_mixer"),
    (("sb2_pipe_outlet", "epnode"), "p_loop_supply_mixer"),
    (("sb3_pipe_outlet", "epnode"), "p_loop_supply_mixer"),
    ("p_loop_demand_mixer", ("db4_pipe_inlet", "epnode")),
    (("db1_pipe_outlet", "epnode"), "p_loop_demand_mixer"),
    (("db2_pipe_outlet", "epnode"), "p_loop_demand_mixer"),
    (("db3_pipe_outlet", "epnode"), "p_loop_demand_mixer"),
]


def nextnode(edges, component):
    """get the next component in the loop"""
    e = edges
    c = component
    n2c = [(a, b) for a, b in e if type(a) == tuple]
    c2nodes = [(a, b) for a, b in e if a == c]
    node2cs = []
    for c2node in c2nodes:
        node2c = [(a, b) for a, b in n2c if a == c2node[-1]]
        if len(node2c) == 0:
            # return []
            node2cs = []
            break
        node2cs.append(node2c[0])
    cs = [b for a, b in node2cs]
    # test for connections that have no nodes
    # filter for no nodes
    nonodes = [(a, b) for a, b in e if type(a) != tuple and type(b) != tuple]
    for a, b in nonodes:
        if a == component:
            cs.append(b)
    return cs


def prevnode(edges, component):
    """get the pervious component in the loop"""
    e = edges
    c = component
    n2c = [(a, b) for a, b in e if type(a) == tuple]
    c2n = [(a, b) for a, b in e if type(b) == tuple]
    node2cs = [(a, b) for a, b in e if b == c]
    c2nodes = []
    for node2c in node2cs:
        c2node = [(a, b) for a, b in c2n if b == node2c[0]]
        if len(c2node) == 0:
            # return []
            c2nodes = []
            break
        c2nodes.append(c2node[0])
    cs = [a for a, b in c2nodes]
    # test for connections that have no nodes
    # filter for no nodes
    nonodes = [(a, b) for a, b in e if type(a) != tuple and type(b) != tuple]
    for a, b in nonodes:
        if b == component:
            cs.append(a)
    return cs


def main():
    edges = e
    # c = 'p_loop_supply_splitter'
    # print next(edges, component)

    c = "Central_Chiller"

    n = 0
    while n == 0:
        print(c)
        nextcs = nextnode(e, c)
        if len(nextcs) == 0:
            break
        c = nextcs[0]
        # next(e, c)

    print("-")

    c = "sb4_pipe"
    n = 0
    while n == 0:
        print(c)
        prevcs = prevnode(e, c)
        if len(prevcs) == 0:
            break
        c = prevcs[0]
        # next(e, c)

    # make diagram
    # from eppy.useful_scripts import loopdiagram1
    # g = loopdiagram1.makediagram(edges)
    # loopdiagram1.save_diagram('a', g)


if __name__ == "__main__":
    main()



# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""pytest for iddgroups"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO
import eppy.EPlusInterfaceFunctions.iddgroups as iddgroups

iddtxt = """!      W/m2, W or deg C
!      W/s
!      W/W
!      years
! **************************************************************************

Lead Input;

Simulation Data;

\\group G1

Version,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier
      \\required-field
      \\default 7.0

Version1,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier
      \\required-field
      \\default 7.0

Version2,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier
      \\required-field
      \\default 7.0

\\group G2

VersionG,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier
      \\required-field
      \\default 7.0

VersionG1,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier
      \\required-field
      \\default 7.0

VersionG2,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier
      \\required-field
      \\default 7.0

"""

idd_commlst = [
    [[]],
    [[]],
    [
        ["unique-object", "format singleLine"],
        ["field Version Identifier", "required-field", "default 7.0"],
    ],
    [
        ["unique-object", "format singleLine"],
        ["field Version Identifier", "required-field", "default 7.0"],
    ],
    [
        ["unique-object", "format singleLine"],
        ["field Version Identifier", "required-field", "default 7.0"],
    ],
    [
        ["unique-object", "format singleLine"],
        ["field Version Identifier", "required-field", "default 7.0"],
    ],
    [
        ["unique-object", "format singleLine"],
        ["field Version Identifier", "required-field", "default 7.0"],
    ],
    [
        ["unique-object", "format singleLine"],
        ["field Version Identifier", "required-field", "default 7.0"],
    ],
]

idf_commdct = [
    [{}],
    [{}],
    [
        {"format": ["singleLine"], "unique-object": [""]},
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
    [
        {"format": ["singleLine"], "unique-object": [""]},
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
    [
        {"format": ["singleLine"], "unique-object": [""]},
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
    [
        {"format": ["singleLine"], "unique-object": [""]},
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
    [
        {"format": ["singleLine"], "unique-object": [""]},
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
    [
        {"format": ["singleLine"], "unique-object": [""]},
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
]


def test_idd2groups():
    """py.test for idd2groups"""
    data = (
        (
            {
                "G2": ["VersionG", "VersionG1", "VersionG2"],
                "G1": ["Version", "Version1", "Version2"],
                None: ["Lead Input", "Simulation Data"],
            },
        ),  # gdict
    )
    for (gdict,) in data:
        result = iddgroups.iddtxt2groups(iddtxt)
        assert result == gdict


def test_idd2group():
    """py.test for idd2group"""
    data = (
        (
            {
                "G2": ["VersionG", "VersionG1", "VersionG2"],
                "G1": ["Version", "Version1", "Version2"],
                None: ["Lead Input", "Simulation Data"],
            },
        ),  # gdict
    )
    for (gdict,) in data:
        fhandle = StringIO(iddtxt)
        result = iddgroups.idd2group(fhandle)
        assert result == gdict


def test_iddtxt2grouplist():
    """py.test for iddtxt2grouplist"""
    data = (
        (
            [
                (None, "Lead Input"),
                (None, "Simulation Data"),
                ("G1", "Version"),
                ("G1", "Version1"),
                ("G1", "Version2"),
                ("G2", "VersionG"),
                ("G2", "VersionG1"),
                ("G2", "VersionG2"),
            ],
        ),  # glist
    )
    for (glist,) in data:
        result = iddgroups.iddtxt2grouplist(iddtxt)
        assert result == glist


def test_idd2grouplist():
    """py.test idd2grouplist"""
    data = (
        (
            [
                (None, "Lead Input"),
                (None, "Simulation Data"),
                ("G1", "Version"),
                ("G1", "Version1"),
                ("G1", "Version2"),
                ("G2", "VersionG"),
                ("G2", "VersionG1"),
                ("G2", "VersionG2"),
            ],
        ),  # glist
    )
    for (glist,) in data:
        fhandle = StringIO(iddtxt)
        result = iddgroups.idd2grouplist(fhandle)
        assert result == glist


def test_group2commlst():
    """py.test for group2commlst"""
    data = (
        (
            [
                [["group None", "idfobj Lead Input"]],
                [["group None", "idfobj Simulation Data"]],
                [
                    [
                        "group G1",
                        "idfobj Version",
                        "unique-object",
                        "format singleLine",
                    ],
                    ["field Version Identifier", "required-field", "default 7.0"],
                ],
                [
                    [
                        "group G1",
                        "idfobj Version1",
                        "unique-object",
                        "format singleLine",
                    ],
                    ["field Version Identifier", "required-field", "default 7.0"],
                ],
                [
                    [
                        "group G1",
                        "idfobj Version2",
                        "unique-object",
                        "format singleLine",
                    ],
                    ["field Version Identifier", "required-field", "default 7.0"],
                ],
                [
                    [
                        "group G2",
                        "idfobj VersionG",
                        "unique-object",
                        "format singleLine",
                    ],
                    ["field Version Identifier", "required-field", "default 7.0"],
                ],
                [
                    [
                        "group G2",
                        "idfobj VersionG1",
                        "unique-object",
                        "format singleLine",
                    ],
                    ["field Version Identifier", "required-field", "default 7.0"],
                ],
                [
                    [
                        "group G2",
                        "idfobj VersionG2",
                        "unique-object",
                        "format singleLine",
                    ],
                    ["field Version Identifier", "required-field", "default 7.0"],
                ],
            ],
        ),  # groupcommlst
    )
    for (groupcommlst,) in data:
        glist = iddgroups.iddtxt2grouplist(iddtxt)
        result = iddgroups.group2commlst(idd_commlst, glist)
        assert result == groupcommlst


(
    (
        [
            (None, "Lead Input"),
            (None, "Simulation Data"),
            ("G1", "Version"),
            ("G1", "Version1"),
            ("G1", "Version2"),
            ("G2", "VersionG"),
            ("G2", "VersionG1"),
            ("G2", "VersionG2"),
        ],
    ),  # glist
)


def test_group2commdct():
    """py.test for group2commdct"""
    data = (
        (
            [
                [{"group": None, "idfobj": "Lead Input"}],
                [{"group": None, "idfobj": "Simulation Data"}],
                [
                    {
                        "group": "G1",
                        "idfobj": "Version",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G1",
                        "idfobj": "Version1",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G1",
                        "idfobj": "Version2",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G2",
                        "idfobj": "VersionG",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G2",
                        "idfobj": "VersionG1",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G2",
                        "idfobj": "VersionG2",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
            ],
        ),  # groupcommdct
    )
    for (groupcommdct,) in data:
        glist = iddgroups.iddtxt2grouplist(iddtxt)
        result = iddgroups.group2commdct(idf_commdct, glist)
        # assert result == groupcommdct
        for r, g in zip(result, groupcommdct):
            assert r == g


def test_commdct2grouplist():
    """py.test for commdct2grouplist"""
    data = (
        (
            [
                [{"group": None, "idfobj": "Lead Input"}],
                [{"group": None, "idfobj": "Simulation Data"}],
                [
                    {
                        "group": "G1",
                        "idfobj": "Version",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G1",
                        "idfobj": "Version1",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G1",
                        "idfobj": "Version2",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G2",
                        "idfobj": "VersionG",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G2",
                        "idfobj": "VersionG1",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
                [
                    {
                        "group": "G2",
                        "idfobj": "VersionG2",
                        "format": ["singleLine"],
                        "unique-object": [""],
                    },
                    {
                        "default": ["7.0"],
                        "field": ["Version Identifier"],
                        "required-field": [""],
                    },
                ],
            ],
            {
                "G2": ["VersionG", "VersionG1", "VersionG2"],
                "G1": ["Version", "Version1", "Version2"],
                None: ["Lead Input", "Simulation Data"],
            },
        ),  # gcommdct, gdict
    )
    for gcommdct, gdict in data:
        result = iddgroups.commdct2grouplist(gcommdct)
        assert result == gdict



# Copyright (c) 2012, 2020 Santosh Philip
# Copyright (c) 2016 Jamie Bull
# Copyright (c) 2020 Cheng Cui
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Sub class Bunch to represent an IDF object.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import itertools

from munch import Munch as Bunch

from eppy.bunchhelpers import matchfieldnames, scientificnotation
import eppy.function_helpers as fh


class BadEPFieldError(AttributeError):
    """An Exception"""

    pass


class RangeError(ValueError):
    """An Exception"""

    pass


def almostequal(first, second, places=7, printit=True):
    """
    Test if two values are equal to a given number of places.
    This is based on python's unittest so may be covered by Python's
    license.

    """
    if first == second:
        return True

    if round(abs(second - first), places) != 0:
        if printit:
            print(round(abs(second - first), places))
            print("notalmost: %s != %s to %i places" % (first, second, places))
        return False
    else:
        return True


def somevalues(ddtt):
    """returns some values"""
    return ddtt.Name, ddtt.Construction_Name, ddtt.obj


def extendlist(lst, i, value=""):
    """extend the list so that you have i-th value"""
    if i < len(lst):
        pass
    else:
        lst.extend([value] * (i - len(lst) + 1))


def return42(self, *args, **kwargs):
    # proof of concept - to be removed
    return 42


def addfunctions(abunch):
    """add functions to epbunch"""

    key = abunch.obj[0].upper()

    # -----------------
    # TODO : alternate strategy to avoid listing the objkeys in snames
    # check if epbunch has field "Zone_Name" or "Building_Surface_Name"
    # and is in group u'Thermal Zones and Surfaces'
    # then it is likely to be a surface.
    # of course we need to recode for surfaces that do not have coordinates :-(
    # or we can filter those out since they do not have
    # the field "Number_of_Vertices"
    snames = [
        "BuildingSurface:Detailed",
        "Wall:Detailed",
        "RoofCeiling:Detailed",
        "Floor:Detailed",
        "FenestrationSurface:Detailed",
        "Shading:Site:Detailed",
        "Shading:Building:Detailed",
        "Shading:Zone:Detailed",
    ]
    snames = [sname.upper() for sname in snames]
    if key in snames:
        func_dict = {
            "area": fh.area,
            "height": fh.height,  # not working correctly
            "width": fh.width,  # not working correctly
            "azimuth": fh.azimuth,
            "true_azimuth": fh.true_azimuth,
            "tilt": fh.tilt,
            "coords": fh.getcoords,  # needed for debugging
        }
        abunch.__functions.update(func_dict)

    # -----------------
    # print(abunch.getfieldidd )
    names = [
        "CONSTRUCTION",
        "MATERIAL",
        "MATERIAL:AIRGAP",
        "MATERIAL:INFRAREDTRANSPARENT",
        "MATERIAL:NOMASS",
        "MATERIAL:ROOFVEGETATION",
        "WINDOWMATERIAL:BLIND",
        "WINDOWMATERIAL:GLAZING",
        "WINDOWMATERIAL:GLAZING:REFRACTIONEXTINCTIONMETHOD",
        "WINDOWMATERIAL:GAP",
        "WINDOWMATERIAL:GAS",
        "WINDOWMATERIAL:GASMIXTURE",
        "WINDOWMATERIAL:GLAZINGGROUP:THERMOCHROMIC",
        "WINDOWMATERIAL:SCREEN",
        "WINDOWMATERIAL:SHADE",
        "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
    ]
    if key in names:
        func_dict = {
            "rvalue": fh.rvalue,
            "ufactor": fh.ufactor,
            "rvalue_ip": fh.rvalue_ip,  # quick fix for Santosh. Needs to thought thru
            "ufactor_ip": fh.ufactor_ip,  # quick fix for Santosh. Needs to thought thru
            "heatcapacity": fh.heatcapacity,
        }
        abunch.__functions.update(func_dict)

    names = [
        "FAN:CONSTANTVOLUME",
        "FAN:VARIABLEVOLUME",
        "FAN:ONOFF",
        "FAN:ZONEEXHAUST",
        "FANPERFORMANCE:NIGHTVENTILATION",
    ]
    if key in names:
        func_dict = {
            "f_fanpower_bhp": fh.fanpower_bhp,
            "f_fanpower_watts": fh.fanpower_watts,
            "f_fan_maxcfm": fh.fan_maxcfm,
        }
        abunch.__functions.update(func_dict)
    # =====
    # code for references
    # -----------------
    # add function zonesurfaces
    if key == "ZONE":
        func_dict = {"zonesurfaces": fh.zonesurfaces}
        abunch.__functions.update(func_dict)

    # -----------------
    # add function subsurfaces
    # going to cheat here a bit
    # check if epbunch has field "Zone_Name"
    # and is in group u'Thermal Zones and Surfaces'
    # then it is likely to be a surface attached to a zone
    fields = abunch.fieldnames
    try:
        group = abunch.getfieldidd("key")["group"]
    except KeyError as e:  # some pytests don't have group
        group = None
    if group == "Thermal Zones and Surfaces":
        if "Zone_Name" in fields:
            func_dict = {"subsurfaces": fh.subsurfaces}
            abunch.__functions.update(func_dict)

    return abunch


class EpBunch(Bunch):
    """
    Fields, values, and descriptions of fields in an EnergyPlus IDF object
    stored in a `bunch` which is a `dict` extended to allow access to dict
    fields as attributes as well as by keys.

    """

    def __init__(self, obj, objls, objidd, *args, **kwargs):
        super(EpBunch, self).__init__(*args, **kwargs)
        self.obj = obj  # field names
        self.objls = objls  # field values
        self.objidd = objidd  # field metadata (minimum, maximum, type, etc.)
        self.theidf = None  # pointer to the idf this epbunch belongs to
        # This is None if there is no idf - a standalone epbunch
        # This will be set by Idf_MSequence
        self["__functions"] = {}  # initialize the functions
        addfunctions(self)

    @property
    def fieldnames(self):
        """Friendly name for objls."""
        return self.objls

    @property
    def fieldvalues(self):
        """Friendly name for obj."""
        return self.obj

    def checkrange(self, fieldname):
        """Check if the value for a field is within the allowed range."""
        return checkrange(self, fieldname)

    def getrange(self, fieldname):
        """Get the allowed range of values for a field."""
        return getrange(self, fieldname)

    def getfieldidd(self, fieldname):
        """get the idd dict for this field
        Will return {} if the fieldname does not exist"""
        return getfieldidd(self, fieldname)

    def getfieldidd_item(self, fieldname, iddkey):
        """return an item from the fieldidd, given the iddkey
        will return and empty list if it does not have the iddkey
        or if the fieldname does not exist"""
        return getfieldidd_item(self, fieldname, iddkey)

    def get_retaincase(self, fieldname):
        """check if the field should retain case"""
        return get_retaincase(self, fieldname)

    def isequal(self, fieldname, value, places=7):
        """return True if the field == value
        Will retain case if get_retaincase == True
        for real value will compare to decimal 'places'
        """
        return isequal(self, fieldname, value, places=places)

    def getreferingobjs(self, iddgroups=None, fields=None):
        """Get a list of objects that refer to this object"""
        return getreferingobjs(self, iddgroups=iddgroups, fields=fields)

    def get_referenced_object(self, fieldname):
        """
        Get an object referred to by a field in another object.

        For example an object of type Construction has fields for each layer, each
        of which refers to a Material. This functions allows the object
        representing a Material to be fetched using the name of the layer.

        Returns the first item found since if there is more than one matching item,
        it is a malformed IDF.

        Parameters
        ----------
        referring_object : EpBunch
            The object which contains a reference to another object,
        fieldname : str
            The name of the field in the referring object which contains the
            reference to another object.

        Returns
        -------
        EpBunch

        """
        return get_referenced_object(self, fieldname)

    def __setattr__(self, name, value):
        try:
            origname = self["__functions"][name]
            # TODO: unit test never hits here so what is it for?
            self[origname] = value
        except KeyError:
            pass

        try:
            name = self["__aliases"][name]  # get original name of the alias
        except KeyError:
            pass

        if name in ("__functions", "__aliases"):  # just set the new value
            self[name] = value
            return None
        elif name in ("obj", "objls", "objidd", "theidf"):  # let Bunch handle it
            super(EpBunch, self).__setattr__(name, value)
            return None
        elif name in self.fieldnames:  # set the value, extending if needed
            i = self.fieldnames.index(name)
            try:
                self.fieldvalues[i] = value
            except IndexError:
                extendlist(self.fieldvalues, i)
                self.fieldvalues[i] = value
        else:
            astr = "unable to find field %s" % (name,)
            raise BadEPFieldError(astr)  # TODO: could raise AttributeError

    def __getattr__(self, name):
        try:
            func = self["__functions"][name]
            return func(self)
        except KeyError:
            pass

        try:
            name = self["__aliases"][name]
        except KeyError:
            pass

        if name == "__functions":
            return self["__functions"]
        elif name in ("__aliases", "obj", "objls", "objidd", "theidf"):
            # unit test
            return super(EpBunch, self).__getattr__(name)
        elif name in self.fieldnames:
            i = self.fieldnames.index(name)
            try:
                return self.fieldvalues[i]
            except IndexError:
                return ""
        else:
            astr = "unable to find field %s" % (name,)
            raise BadEPFieldError(astr)

    def __getitem__(self, key):
        if key in ("obj", "objls", "objidd", "__functions", "__aliases", "theidf"):
            return super(EpBunch, self).__getitem__(key)
        elif key in self.fieldnames:
            i = self.fieldnames.index(key)
            try:
                return self.fieldvalues[i]
            except IndexError:
                return ""
        else:
            astr = "unknown field %s" % (key,)
            raise BadEPFieldError(astr)

    def __setitem__(self, key, value):
        if key in ("obj", "objls", "objidd", "__functions", "__aliases", "theidf"):
            super(EpBunch, self).__setitem__(key, value)
            return None
        elif key in self.fieldnames:
            i = self.fieldnames.index(key)
            try:
                self.fieldvalues[i] = value
            except IndexError:
                extendlist(self.fieldvalues, i)
                self.fieldvalues[i] = value
        else:
            astr = "unknown field %s" % (key,)
            raise BadEPFieldError(astr)

    def __repr__(self):
        """print this as an idf snippet"""
        # lines = [str(val) for val in self.obj]
        # replace the above line with code that will print an integer without decimals
        lines = []
        for val in self.obj:
            try:
                value = int(val)
                if value != val:
                    value = val
            except ValueError as e:
                value = val
            lines.append(value)
        # ------------
        comments = [comm.replace("_", " ") for comm in self.objls]
        lines[0] = "%s," % (lines[0],)  # comma after first line
        for i, line in enumerate(lines[1:-1]):
            line = scientificnotation(
                line, width=18
            )  # E+ cannot read wide numbers, convert to 1e+3
            lines[i + 1] = "    %s," % (line,)  # indent and comma
        lines[-1] = "    %s;" % (lines[-1],)  # ';' after last line
        lines = lines[:1] + [line.ljust(26) for line in lines[1:]]  # ljsut the lines
        filler = "%s    !- %s"
        nlines = [
            filler % (line, comm) for line, comm in zip(lines[1:], comments[1:])
        ]  # adds comments to line
        nlines.insert(0, lines[0])  # first line without comment
        astr = "\n".join(nlines)
        return "\n%s\n" % (astr,)

    def __str__(self):
        """same as __repr__"""
        # needed if YAML is installed. See issue 67
        # unit test
        return self.__repr__()

    def __dir__(self):
        fnames = self.fieldnames
        func_names = list(self["__functions"].keys())
        return super(EpBunch, self).__dir__() + fnames + func_names


def getrange(bch, fieldname):
    """get the ranges for this field"""
    keys = ["maximum", "minimum", "maximum<", "minimum>", "type"]
    index = bch.objls.index(fieldname)
    fielddct_orig = bch.objidd[index]
    fielddct = copy.deepcopy(fielddct_orig)
    therange = {}
    for key in keys:
        therange[key] = fielddct.setdefault(key, None)
    if therange["type"]:
        therange["type"] = therange["type"][0]
    if therange["type"] == "real":
        for key in keys[:-1]:
            if therange[key]:
                therange[key] = float(therange[key][0])
    if therange["type"] == "integer":
        for key in keys[:-1]:
            if therange[key]:
                therange[key] = int(therange[key][0])
    return therange


def checkrange(bch, fieldname):
    """throw exception if the out of range"""
    fieldvalue = bch[fieldname]
    therange = bch.getrange(fieldname)
    if therange["maximum"] != None:
        if fieldvalue > therange["maximum"]:
            astr = "Value %s is not less or equal to the 'maximum' of %s"
            astr = astr % (fieldvalue, therange["maximum"])
            raise RangeError(astr)
    if therange["minimum"] != None:
        if fieldvalue < therange["minimum"]:
            astr = "Value %s is not greater or equal to the 'minimum' of %s"
            astr = astr % (fieldvalue, therange["minimum"])
            raise RangeError(astr)
    if therange["maximum<"] != None:
        if fieldvalue >= therange["maximum<"]:
            astr = "Value %s is not less than the 'maximum<' of %s"
            astr = astr % (fieldvalue, therange["maximum<"])
            raise RangeError(astr)
    if therange["minimum>"] != None:
        if fieldvalue <= therange["minimum>"]:
            astr = "Value %s is not greater than the 'minimum>' of %s"
            astr = astr % (fieldvalue, therange["minimum>"])
            raise RangeError(astr)
    return fieldvalue
    """get the idd dict for this field
    Will return {} if the fieldname does not exist"""


def getfieldidd(bch, fieldname):
    """get the idd dict for this field
    Will return {} if the fieldname does not exist"""
    # print(bch)
    try:
        fieldindex = bch.objls.index(fieldname)
    except ValueError as e:
        return {}  # the fieldname does not exist
        # so there is no idd
    fieldidd = bch.objidd[fieldindex]
    return fieldidd


def getfieldidd_item(bch, fieldname, iddkey):
    """return an item from the fieldidd, given the iddkey
    will return and empty list if it does not have the iddkey
    or if the fieldname does not exist"""
    fieldidd = getfieldidd(bch, fieldname)
    try:
        return fieldidd[iddkey]
    except KeyError as e:
        return []


def get_retaincase(bch, fieldname):
    """Check if the field should retain case"""
    fieldidd = bch.getfieldidd(fieldname)
    return "retaincase" in fieldidd


def isequal(bch, fieldname, value, places=7):
    """return True if the field is equal to value"""

    def equalalphanumeric(bch, fieldname, value):
        if bch.get_retaincase(fieldname):
            return bch[fieldname] == value
        else:
            return bch[fieldname].upper() == value.upper()

    fieldidd = bch.getfieldidd(fieldname)
    try:
        ftype = fieldidd["type"][0]
        if ftype in ["real", "integer"]:
            return almostequal(bch[fieldname], float(value), places=places)
        else:
            return equalalphanumeric(bch, fieldname, value)
    except KeyError as e:
        return equalalphanumeric(bch, fieldname, value)


def getreferingobjs(referedobj, iddgroups=None, fields=None):
    """Get a list of objects that refer to this object"""
    # pseudocode for code below
    # referringobjs = []
    # referedobj has: -> Name
    #                 -> reference
    # for each obj in idf:
    # [optional filter -> objects in iddgroup]
    #     each field of obj:
    #     [optional filter -> field in fields]
    #         has object-list [refname]:
    #             if refname in reference:
    #                 if Name = field value:
    #                     referringobjs.append()
    referringobjs = []
    idf = referedobj.theidf
    referedidd = referedobj.getfieldidd("Name")
    try:
        references = referedidd["reference"]
    except KeyError as e:
        return referringobjs
    idfobjs = idf.idfobjects.values()
    idfobjs = list(itertools.chain.from_iterable(idfobjs))  # flatten list
    if iddgroups:  # optional filter
        idfobjs = [
            anobj for anobj in idfobjs if anobj.getfieldidd("key")["group"] in iddgroups
        ]
    for anobj in idfobjs:
        if not fields:
            thefields = anobj.objls
        else:
            thefields = fields
        for field in thefields:
            try:
                itsidd = anobj.getfieldidd(field)
            except ValueError as e:
                continue
            if "object-list" in itsidd:
                refname = itsidd["object-list"][0]
                if refname in references:
                    if referedobj.isequal("Name", anobj[field]):
                        referringobjs.append(anobj)
    return referringobjs


def get_referenced_object(referring_object, fieldname):
    """
    Get an object referred to by a field in another object.

    For example an object of type Construction has fields for each layer, each
    of which refers to a Material. This functions allows the object
    representing a Material to be fetched using the name of the layer.

    Returns the first item found since if there is more than one matching item,
    it is a malformed IDF.

    Parameters
    ----------
    referring_object : EpBunch
        The object which contains a reference to another object,
    fieldname : str
        The name of the field in the referring object which contains the
        reference to another object.

    Returns
    -------
    EpBunch

    """
    idf = referring_object.theidf
    object_list = referring_object.getfieldidd_item(fieldname, "object-list")
    for obj_type in idf.idfobjects:
        for obj in idf.idfobjects[obj_type]:
            valid_object_lists = obj.getfieldidd_item("Name", "reference")
            if set(object_list).intersection(set(valid_object_lists)):
                referenced_obj_name = referring_object[fieldname]
                if obj.Name == referenced_obj_name:
                    return obj



# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""module for idd_index data structure and functions to work with it
- idd_index indexes idd_info so that it is easy to search through it.
- idd_info is the datastructure that holds the info in the Energy+.idd file"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


# possible data structure:
# idd_index = {
#     ref2names:{
#         MaterialName:[MATERIAL, MATERIAL:AIRGAP, ...]
#         ZoneName:[ZONE,]
#     },
#     name2refs:{
#         ZONE:[ZoneNames, OutFaceEnvNames, ...]
#         SCHEDULE:DAY:HOURLY:[DayScheduleNames, ScheduleAndDayScheduleNames]
#     },
#     fieldsthatrefer:{
#         WINDOW:[], # will be list of fields in the future
#         DOOR:[], # once I figure how to not list all the repeating fields
#     }
#
# }

# embedd (actually point to) some of this info in idd_info
# In [17]: construction.getfieldidd('Layer_2')
# Out[17]:
# {u'field': [u'Layer 2'],
#  u'object-list': [u'MaterialName'],
#  u'type': [u'object-list']}
# will now become:
# Out[17]:
# {u'field': [u'Layer 2'],
#  u'object-list': [u'MaterialName'],
#  'referedobjkeys':[MATERIAL, MATERIAL:AIRGAP, ...]
#  # MaterialName:[MATERIAL, MATERIAL:AIRGAP, ...] is the same list as in
#  # ref2names above, not a new list. So we don't use too much memory
#  u'type': [u'object-list']}

from eppy.bunch_subclass import getfieldidd_item


def makename2refdct(commdct):
    """make the name2refs dict in the idd_index"""
    refdct = {}
    for comm in commdct:  # commdct is a list of dict
        try:
            idfobj = comm[0]["idfobj"].upper()
            field1 = comm[1]
            if "Name" in field1["field"]:
                references = field1["reference"]
                refdct[idfobj] = references
        except (KeyError, IndexError) as e:
            continue  # not the expected pattern for reference
    return refdct


def makeref2namesdct(name2refdct):
    """make the ref2namesdct in the idd_index"""
    ref2namesdct = {}
    for key, values in name2refdct.items():
        for value in values:
            ref2namesdct.setdefault(value, set()).add(key)
    return ref2namesdct


def ref2names2commdct(ref2names, commdct):
    """embed ref2names into commdct"""
    for comm in commdct:
        for cdct in comm:
            try:
                refs = cdct["object-list"][0]
                validobjects = ref2names[refs]
                cdct.update({"validobjects": validobjects})
            except KeyError as e:
                continue
    return commdct



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""pytest for idfreader. very few tests"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO

import eppy.idfreader as idfreader
from eppy.EPlusInterfaceFunctions import readidf

from eppy.iddcurrent import iddcurrent

iddfhandle = StringIO(iddcurrent.iddtxt)


def test_iddversiontuple():
    """py.test for iddversiontuple"""
    iddtxt = """stuff 9.8.4
    other stuff"""
    fhandle = StringIO(iddtxt)
    result = idfreader.iddversiontuple(fhandle)
    assert result == (9, 8, 4)


def test_convertafield():
    """py.test for convertafield"""
    data = (
        ({"type": "integer"}, "1", "N1", 1),
        # field_comm, field_val, field_iddname, expected
        ({}, "1", "N1", 1.0),
        # field_comm, field_val, field_iddname, expected
        ({"type": "real"}, "1", "N1", 1.0),
        # field_comm, field_val, field_iddname, expected
        ({}, "autosize", "N1", "autosize"),
        ({}, "4", "A1", "4"),
        # field_comm, field_val, field_iddname, expected
    )
    for field_comm, field_val, field_iddname, expected in data:
        result = idfreader.convertafield(field_comm, field_val, field_iddname)
        assert result == expected


def test_convertallfields():
    """py.test convertallfields"""
    data = (
        ("version, 8.1;", "VERSION", ["version", "8.1"]),
        # idfstr, objkey, expected
        (
            "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM, simple, 0.45;",
            "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
            ["WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM", "simple", 0.45],
        ),
        # idfstr, objkey, expected
        (
            "HVACTEMPLATE:ZONE:FANCOIL, gumby1, gumby2, 0.45;",
            "HVACTEMPLATE:ZONE:FANCOIL",
            ["HVACTEMPLATE:ZONE:FANCOIL", "gumby1", "gumby2", 0.45],
        ),
        # idfstr, objkey, expected
        (
            "HVACTEMPLATE:ZONE:FANCOIL, gumby1, gumby2, autosize;",
            "HVACTEMPLATE:ZONE:FANCOIL",
            ["HVACTEMPLATE:ZONE:FANCOIL", "gumby1", "gumby2", "autosize"],
        ),
        # idfstr, objkey, expected
    )
    commdct = None
    block = None
    for idfstr, objkey, expected in data:
        idfhandle = StringIO(idfstr)
        block, data, commdct, idd_index = readidf.readdatacommdct1(
            idfhandle, iddfile=iddfhandle, commdct=commdct, block=block
        )
        idfreader.convertallfields(data, commdct, block)
        result = data.dt[objkey][0]
        assert result == expected


def test_getextensible():
    """py.test for getextensible"""
    data = (
        (
            [
                {
                    "format": ["singleLine"],
                    "group": "Simulation Parameters",
                    "idfobj": "Version",
                    "memo": ["Specifies the EnergyPlus version of the IDF file."],
                    "unique-object": [""],
                },
                {},
                {},
                {},
            ],
            None,
        ),  # objidd, expected
        (
            [
                {
                    "extensible:2": [
                        '- repeat last two fields, remembering to remove ; from "inner" fields.'
                    ],
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                    "memo": [
                        "A Schedule:Day:Interval contains a full day of values with specified end times for each value",
                        "Currently, is set up to allow for 10 minute intervals for an entire day.",
                    ],
                    "min-fields": ["5"],
                },
                {},
                {},
                {},
            ],
            2,
        ),  # objidd, expected
    )
    for objidd, expected in data:
        result = idfreader.getextensible(objidd)
        assert result == expected


def test_endof_extensible():
    """py.test for endof_extensible"""
    data = (
        (1, ["gumby", "A1", "A2"], ["A2"]),  # extensible, thisblock, expected
        (1, ["gumby", "A1", "A2", "A3"], ["A3"]),  # extensible, thisblock, expected
        (
            2,
            ["gumby", "A1", "A2", "A3"],
            ["A2", "A3"],
        ),  # extensible, thisblock, expected
    )
    for extensible, thisblock, expected in data:
        result = idfreader.endof_extensible(extensible, thisblock)
        assert result == expected


def test_extension_of_extensible():
    """py.test for extension_of_extensible"""
    data = (
        (
            [
                {
                    "extensible:1": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            1,
            ["A3"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:1": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            2,
            ["A3", "A4"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:1": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            3,
            ["A3", "A4", "A5"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:2": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            2,
            ["A3", "A4"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:4": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "N3", "A4", "M8", "A5"],
            4,
            ["N4", "A6", "M9", "A7"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:4": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "N3", "A4", "M8", "A5"],
            12,
            [
                "N4",
                "A6",
                "M9",
                "A7",
                "N5",
                "A8",
                "M10",
                "A9",
                "N6",
                "A10",
                "M11",
                "A11",
            ],
        ),  # objidd, objblock, n, expected
    )
    for objidd, objblock, n, expected in data:
        result = idfreader.extension_of_extensible(objidd, objblock, n)
        assert result == expected



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""get the loop data to draw the diagram
Other notes:
- tested for idd version 6.0
- when E+ is updated, run versionchangecheck.py for the following objects
uses the following objects
['plantloop', ]
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


def extractfields(data, commdct, objkey, fieldlists):
    """get all the objects of objkey.
    fieldlists will have a fieldlist for each of those objects.
    return the contents of those fields"""
    # TODO : this assumes that the field list identical for
    # each instance of the object. This is not true.
    # So we should have a field list for each instance of the object
    # and map them with a zip
    objindex = data.dtls.index(objkey)
    objcomm = commdct[objindex]
    objfields = []
    # get the field names of that object
    for dct in objcomm[0:]:
        try:
            thefieldcomms = dct["field"]
            objfields.append(thefieldcomms[0])
        except KeyError as err:
            objfields.append(None)
    fieldindexes = []
    for fieldlist in fieldlists:
        fieldindex = []
        for item in fieldlist:
            if isinstance(item, int):
                fieldindex.append(item)
            else:
                fieldindex.append(objfields.index(item) + 0)
                # the index starts at 1, not at 0
        fieldindexes.append(fieldindex)
    theobjects = data.dt[objkey]
    fieldcontents = []
    for theobject, fieldindex in zip(theobjects, fieldindexes):
        innerlst = []
        for item in fieldindex:
            try:
                innerlst.append(theobject[item])
            except IndexError as err:
                break
        fieldcontents.append(innerlst)
        # fieldcontents.append([theobject[item] for item in fieldindex])
    return fieldcontents


def plantloopfieldlists(data):
    """return the plantloopfield list"""
    objkey = "plantloop".upper()
    numobjects = len(data.dt[objkey])
    return [
        [
            "Name",
            "Plant Side Inlet Node Name",
            "Plant Side Outlet Node Name",
            "Plant Side Branch List Name",
            "Demand Side Inlet Node Name",
            "Demand Side Outlet Node Name",
            "Demand Side Branch List Name",
        ]
    ] * numobjects


def plantloopfields(data, commdct):
    """get plantloop fields to diagram it"""
    fieldlists = plantloopfieldlists(data)
    objkey = "plantloop".upper()
    return extractfields(data, commdct, objkey, fieldlists)


def branchlist2branches(data, commdct, branchlist):
    """get branches from the branchlist"""
    objkey = "BranchList".upper()
    theobjects = data.dt[objkey]
    fieldlists = []
    objnames = [obj[1] for obj in theobjects]
    for theobject in theobjects:
        fieldlists.append(list(range(2, len(theobject))))
    blists = extractfields(data, commdct, objkey, fieldlists)
    thebranches = [
        branches for name, branches in zip(objnames, blists) if name == branchlist
    ]
    return thebranches[0]


def branch_inlet_outlet(data, commdct, branchname):
    """return the inlet and outlet of a branch"""
    objkey = "Branch".upper()
    theobjects = data.dt[objkey]
    theobject = [obj for obj in theobjects if obj[1] == branchname]
    theobject = theobject[0]
    inletindex = 6
    outletindex = len(theobject) - 2
    return [theobject[inletindex], theobject[outletindex]]


def splittermixerfieldlists(data, commdct, objkey):
    """docstring for splittermixerfieldlists"""
    objkey = objkey.upper()
    objindex = data.dtls.index(objkey)
    objcomms = commdct[objindex]
    theobjects = data.dt[objkey]
    fieldlists = []
    for theobject in theobjects:
        fieldlist = list(range(1, len(theobject)))
        fieldlists.append(fieldlist)
    return fieldlists


def splitterfields(data, commdct):
    """get splitter fields to diagram it"""
    objkey = "Connector:Splitter".upper()
    fieldlists = splittermixerfieldlists(data, commdct, objkey)
    return extractfields(data, commdct, objkey, fieldlists)


def mixerfields(data, commdct):
    """get mixer fields to diagram it"""
    objkey = "Connector:Mixer".upper()
    fieldlists = splittermixerfieldlists(data, commdct, objkey)
    return extractfields(data, commdct, objkey, fieldlists)


def repeatingfields(theidd, commdct, objkey, flds):
    """return a list of repeating fields
    fld is in format 'Component %s Name'
    so flds = [fld % (i, ) for i in range(n)]
    does not work for 'fields as indicated'"""
    # TODO : make it work for 'fields as indicated'
    if type(flds) != list:
        flds = [flds]  # for backward compatability
    objindex = theidd.dtls.index(objkey)
    objcomm = commdct[objindex]
    allfields = []
    for fld in flds:
        thefields = []
        indx = 1
        for i in range(len(objcomm)):
            try:
                thefield = fld % (indx,)
                if objcomm[i]["field"][0] == thefield:
                    thefields.append(thefield)
                    indx = indx + 1
            except KeyError as err:
                pass
        allfields.append(thefields)
    allfields = list(zip(*allfields))
    return [item for sublist in allfields for item in sublist]


def objectcount(data, key):
    """return the count of objects of key"""
    objkey = key.upper()
    return len(data.dt[objkey])


def getfieldindex(data, commdct, objkey, fname):
    """given objkey and fieldname, return its index"""
    objindex = data.dtls.index(objkey)
    objcomm = commdct[objindex]
    for i_index, item in enumerate(objcomm):
        try:
            if item["field"] == [fname]:
                break
        except KeyError as err:
            pass
    return i_index


def getadistus(data, commdct):
    """docstring for fname"""
    objkey = "ZoneHVAC:AirDistributionUnit".upper()
    objindex = data.dtls.index(objkey)
    objcomm = commdct[objindex]
    adistutypefield = "Air Terminal Object Type"
    ifield = getfieldindex(data, commdct, objkey, adistutypefield)
    adistus = objcomm[ifield]["key"]
    return adistus


def makeadistu_inlets(data, commdct):
    """make the dict adistu_inlets"""
    adistus = getadistus(data, commdct)
    # assume that the inlet node has the words "Air Inlet Node Name"
    airinletnode = "Air Inlet Node Name"
    adistu_inlets = {}
    for adistu in adistus:
        objkey = adistu.upper()
        objindex = data.dtls.index(objkey)
        objcomm = commdct[objindex]
        airinlets = []
        for i, comm in enumerate(objcomm):
            try:
                if comm["field"][0].find(airinletnode) != -1:
                    airinlets.append(comm["field"][0])
            except KeyError as err:
                pass
        adistu_inlets[adistu] = airinlets
    return adistu_inlets



# Copyright (c) 2017 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Tests for fanpower.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO

import eppy.fanpower as fanpower
from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF
from eppy.pytest_helpers import almostequal


iddfhandle = StringIO(iddcurrent.iddtxt)

if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)


def test_pascal2inh2o():
    """py.test for pascal2inh2o and inh2o2pascal"""
    data = (
        (1, 0.00401865),  # pascal, expected
        (186.816675, 0.7507501808391),  # pascal, expected
    )
    for pascal, expected in data:
        result = fanpower.pascal2inh2o(pascal)
        assert almostequal(result, expected, places=5)
    data = (
        (1, 0.00401865),  # expected, inh20
        (186.816675, 0.7507501808391),  # expected, inh20
    )
    for expected, inh20 in data:
        result = fanpower.inh2o2pascal(inh20)
        assert almostequal(result, expected, places=3)


def test_m3s2cfm():
    """py.test for m3s2cfm and cfm2m3s"""
    data = (
        (1, 2118.880003),  # m3s, expected
        (1.28442384, 2721.539989952472),  # m3s, expected
    )
    for m3s, expected in data:
        result = fanpower.m3s2cfm(m3s)
        assert result == expected
    data = (
        (1, 2118.880003),  # expected, cfm
        (1.28442384, 2721.539989952472),  # expected, cfm
    )
    for expected, cfm in data:
        result = fanpower.cfm2m3s(cfm)
        assert result == expected


def test_fan_bhp():
    """py.test for fan_bhp"""
    data = (
        (
            0.342,
            186.816675,
            1.28442384,
            0.939940898974,
        ),  # fan_tot_eff, pascal, m3s, expected
        (
            0.537,
            871.81115,
            2.44326719,
            5.31400249068,
        ),  # fan_tot_eff, pascal, m3s, expected
    )
    for fan_tot_eff, pascal, m3s, expected in data:
        result = fanpower.fan_bhp(fan_tot_eff, pascal, m3s)
        assert almostequal(result, expected)


def test_bhp2watts():
    """py.test for bhp2watts and watts2bhp"""
    data = (
        (0.939940898974, 700.9139283649118),  # bhp, expected
        (5.31400249068, 3962.6516573000763),  # bhp, expected
    )
    for bhp, expected in data:
        result = fanpower.bhp2watts(bhp)
        assert result == expected
    data = (
        (0.939940898974, 700.9139283649118),  # expected, watts
        (5.31400249068, 3962.6516573000763),  # expected, watts
    )
    for expected, watts in data:
        result = fanpower.watts2bhp(watts)
        assert result == expected


def test_fan_watts():
    """py.test for fan_watts"""
    data = (
        (
            0.342,
            186.816675,
            1.28442384,
            700.9139283649118,
        ),  # fan_tot_eff, pascal, m3s, expected
        (
            0.537,
            871.81115,
            2.44326719,
            3962.6516573000763,
        ),  # fan_tot_eff, pascal, m3s, expected
    )
    for fan_tot_eff, pascal, m3s, expected in data:
        result = fanpower.fan_watts(fan_tot_eff, pascal, m3s)
        assert almostequal(result, expected)


vavfan = """Fan:VariableVolume,
  Fan 2,                                  !- Name
  OfficeHVACAvail,                        !- Availability Schedule Name
  0.519,                                  !- Fan Total Efficiency
  164.3824832215,                         !- Pressure Rise {Pa}
  5.6633693184,                           !- Maximum Flow Rate {m3/s}
  FixedFlowRate,                          !- Fan Power Minimum Flow Rate Input Method
  0,                                      !- Fan Power Minimum Flow Fraction
  1.4158423296,                           !- Fan Power Minimum Air Flow Rate {m3/s}
  0.865,                                  !- Motor Efficiency
  1,                                      !- Motor In Airstream Fraction
  0.04076,                                !- Fan Power Coefficient 1
  0.088045,                               !- Fan Power Coefficient 2
  -0.072926,                              !- Fan Power Coefficient 3
  0.94374,                                !- Fan Power Coefficient 4
  0,                                      !- Fan Power Coefficient 5
  AC(1)(2) Supply Side (Return Air) Inlet Node, !- Air Inlet Node Name
  Fan 2 Outlet Node;                      !- Air Outlet Node Name
"""


def testfanpower_bhp():
    """py.test for fanpower_bhp in idf"""
    idf = IDF(StringIO(vavfan))
    thefans = idf.idfobjects["Fan:VariableVolume".upper()]
    thefan = thefans[0]
    bhp = thefan.f_fanpower_bhp
    assert almostequal(bhp, 2.40306611606)
    # test autosize
    thefan.Maximum_Flow_Rate = "autosize"
    bhp = thefan.f_fanpower_bhp
    assert bhp == "autosize"


def testfanpower_watts():
    """py.test for fanpower_watts in idf"""
    idf = IDF(StringIO(vavfan))
    thefans = idf.idfobjects["Fan:VariableVolume".upper()]
    thefan = thefans[0]
    watts = thefan.f_fanpower_watts
    assert almostequal(watts, 1791.9664027495671)
    # test autosize
    thefan.Maximum_Flow_Rate = "autosize"
    watts = thefan.f_fanpower_watts
    assert watts == "autosize"


def test_fan_maxcfm():
    """py.test for fan_maxcfm in idf"""
    idf = IDF(StringIO(vavfan))
    thefans = idf.idfobjects["Fan:VariableVolume".upper()]
    thefan = thefans[0]
    cfm = thefan.f_fan_maxcfm
    assert almostequal(cfm, 12000, places=5)
    # test autosize
    thefan.Maximum_Flow_Rate = "autosize"
    watts = thefan.f_fanpower_watts
    assert watts == "autosize"


def test_bhp2pascal():
    """py.test for bhp2pascal"""
    bhp = 10.182489271962908
    cfm = 74999.99998975938
    fan_tot_eff = 0.58
    exp_pascal, exp_m3s = (124.544455, 35.39605824)
    result_pascal, result_m3s = fanpower.bhp2pascal(bhp, cfm, fan_tot_eff)
    assert almostequal(result_pascal, exp_pascal)
    assert almostequal(result_m3s, exp_m3s)


def test_watts2pascal():
    """py.test for watts2pascal"""
    watts = 7593.0822501027405
    cfm = 74999.99998975938
    fan_tot_eff = 0.58
    exp_pascal, exp_m3s = (124.544455, 35.39605824)
    result_pascal, result_m3s = fanpower.watts2pascal(watts, cfm, fan_tot_eff)
    assert almostequal(result_pascal, exp_pascal)
    assert almostequal(result_m3s, exp_m3s)



import os

import pytest
from io import StringIO
from importlib import reload

from eppy.modeleditor import IDF
from eppy.iddcurrent import iddcurrent
from eppy import modeleditor
from .test_runner import versiontuple

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

RESOURCES_DIR = os.path.join(THIS_DIR, os.pardir, "eppy", "resources")

IDD_FILES = os.path.join(RESOURCES_DIR, "iddfiles")
IDF_FILES = os.path.join(RESOURCES_DIR, "idffiles")
try:
    VERSION = os.environ["ENERGYPLUS_INSTALL_VERSION"]  # used in CI files
except KeyError:
    VERSION = "8-9-0"  # current default for integration tests on local system
TEST_IDF = "V{}/smallfile.idf".format(VERSION[:3].replace("-", "_"))
TEST_EPW = "USA_CA_San.Francisco.Intl.AP.724940_TMY3.epw"
TEST_IDD = "Energy+V{}.idd".format(VERSION.replace("-", "_"))
TEST_OLD_IDD = "Energy+V7_2_0.idd"


@pytest.fixture()
def test_idf():
    idd_file = os.path.join(IDD_FILES, TEST_IDD)
    idf_file = os.path.join(IDF_FILES, TEST_IDF)
    reload(modeleditor)
    modeleditor.IDF.setiddname(idd_file, testing=True)
    idf = modeleditor.IDF(idf_file, TEST_EPW)
    try:
        ep_version = idf.idd_version
        assert ep_version == versiontuple(VERSION)
    except AttributeError:
        raise
    return idf


@pytest.fixture()
def base_idf():
    iddsnippet = iddcurrent.iddtxt
    iddfhandle = StringIO(iddsnippet)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)
    idftxt = ""
    idfhandle = StringIO(idftxt)
    idf = IDF(idfhandle)
    return idf



# Copyright (c) 2020 Cheng Cui
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.test for function_helpers"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO
import pytest

import eppy.function_helpers as fh
from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF
from eppy.pytest_helpers import almostequal

iddtxt = iddcurrent.iddtxt
iddfhandle = StringIO(iddcurrent.iddtxt)
if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)

idftxt = """Version,8.0;

    Building,
        Simple One Zone,         !- Name
        ;                        !- North Axis {deg} 

    Zone,
        ZONE ONE,                !- Name
        ,                        !- Direction of Relative North {deg}
        0, 0, 0;                            !- X,Y,Z  {m}

    GlobalGeometryRules,
        UpperLeftCorner,         !- Starting Vertex Position
        CounterClockWise,        !- Vertex Entry Direction
        World;                   !- Coordinate System

    BuildingSurface:Detailed,
        Zn001:Wall001,           !- Name
        Wall,                    !- Surface Type
        R13WALL,                 !- Construction Name
        ZONE ONE,                !- Zone Name
        Outdoors,                !- Outside Boundary Condition
        ,                        !- Outside Boundary Condition Object
        SunExposed,              !- Sun Exposure
        WindExposed,             !- Wind Exposure
        0.5000000,               !- View Factor to Ground
        4,                       !- Number of Vertices
        0, 0, 4.572000,                     !- X,Y,Z  1 {m}
        0, 0, 0,                            !- X,Y,Z  2 {m}
        15.24000, 0, 0,                     !- X,Y,Z  3 {m}
        15.24000, 0, 4.572000;              !- X,Y,Z  4 {m}

"""


# original test_true_azimuth() - done using a loop
# def test_true_azimuth():
#     """py.test for true_azimuth"""
#     data = (
#         ("Relative", 45, 30, 180 + 75),
#         # coord_system, bldg_north, zone_rel_north, expected,
#         ("Relative", "", 0, 180 + 0),
#         ("World", 45, "", 180),
#         ("Relative", 240, 90, 180 + 330 - 360),
#     )
#
#     fhandle = StringIO(idftxt)
#     idf = IDF(fhandle)
#     geom_rules = idf.idfobjects["GlobalGeometryRules"][0]
#     building = idf.idfobjects["Building"][0]
#     zone = idf.idfobjects["Zone"][0]
#     surface = idf.idfobjects["BuildingSurface:Detailed"][0]
#
#     for coord_system, bldg_north, zone_rel_north, expected in data:
#         geom_rules.Coordinate_System = coord_system
#         building.North_Axis = bldg_north
#         zone.Direction_of_Relative_North = zone_rel_north
#         result = fh.true_azimuth(surface)
#         assert almostequal(expected, result, places=3) == True

# test_true_azimuth() - done using @pytest.mark.parametrize
# see https://docs.pytest.org/en/stable/parametrize.html
@pytest.mark.parametrize(
    "coord_system, bldg_north, zone_rel_north, expected",
    [
        ("Relative", 45, 30, 180 + 75),
        # coord_system, bldg_north, zone_rel_north, expected,
        ("Relative", "", 0, 180 + 0),
        ("World", 45, "", 180),
        ("Relative", 240, 90, 180 + 330 - 360),
    ],
)
def test_true_azimuth(coord_system, bldg_north, zone_rel_north, expected):
    """py.test for true_azimuth"""
    fhandle = StringIO(idftxt)
    idf = IDF(fhandle)
    geom_rules = idf.idfobjects["GlobalGeometryRules"][0]
    building = idf.idfobjects["Building"][0]
    zone = idf.idfobjects["Zone"][0]
    surface = idf.idfobjects["BuildingSurface:Detailed"][0]

    geom_rules.Coordinate_System = coord_system
    building.North_Axis = bldg_north
    zone.Direction_of_Relative_North = zone_rel_north
    result = fh.true_azimuth(surface)
    assert almostequal(expected, result, places=3) == True


def test_true_azimuth_exception():
    """py.test for true_azimuth exception"""
    coord_system, bldg_north, zone_rel_north = (
        "Global",
        0,
        0,
    )

    fhandle = StringIO(idftxt)
    idf = IDF(fhandle)
    geom_rules = idf.idfobjects["GlobalGeometryRules"][0]
    building = idf.idfobjects["Building"][0]
    zone = idf.idfobjects["Zone"][0]
    surface = idf.idfobjects["BuildingSurface:Detailed"][0]

    geom_rules.Coordinate_System = coord_system
    building.North_Axis = bldg_north
    zone.Direction_of_Relative_North = zone_rel_north

    with pytest.raises(ValueError):
        result = fh.true_azimuth(surface)



# -*- coding: utf-8 -*-
# Copyright (c) 2017 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""quick and dirty functions for get fan power BHP or Watts
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# from eppy.bunch_subclass import BadEPFieldError
# import eppy.bunch_subclass


def pascal2inh2o(pascal):
    """convert pressure in Pascals to inches of water"""
    # got this from a google search
    return pascal * 0.00401865


def inh2o2pascal(inh2o):
    """convert pressure in inches of water to Pascals"""
    return (inh2o * 1.0) / 0.00401865


def m3s2cfm(m3s):
    """convert flow meter^3/second to cfm"""
    # from http://www.traditionaloven.com/tutorials/flow-rate/convert-m3-cubic-meter-per-second-to-ft3-cubic-foot-per-minute.html
    return (m3s * 1.0) * 2118.880003


def cfm2m3s(cfm):
    """convert flow cfm to meter^3/second"""
    return cfm / 2118.880003


def fan_bhp(fan_tot_eff, pascal, m3s):
    """return the fan power in bhp given fan efficiency, Pressure rise (Pa) and flow (m3/s)"""
    # from discussion in
    # http://energy-models.com/forum/baseline-fan-power-calculation
    inh2o = pascal2inh2o(pascal)
    cfm = m3s2cfm(m3s)
    return (cfm * inh2o * 1.0) / (6356.0 * fan_tot_eff)


def bhp2pascal(bhp, cfm, fan_tot_eff):
    """return inputs for E+ in pascal and m3/s"""
    inh2o = bhp * 6356.0 * fan_tot_eff / cfm
    pascal = inh2o2pascal(inh2o)
    m3s = cfm2m3s(cfm)
    return pascal, m3s


def bhp2watts(bhp):
    """convert brake horsepower (bhp) to watts"""
    return bhp * 745.7


def watts2bhp(watts):
    """convert watts to brake horsepower (bhp)"""
    return watts / 745.7


def fan_watts(fan_tot_eff, pascal, m3s):
    """return the fan power in watts given fan efficiency, Pressure rise (Pa) and flow (m3/s)"""
    # got this from a google search
    bhp = fan_bhp(fan_tot_eff, pascal, m3s)
    return bhp2watts(bhp)


def watts2pascal(watts, cfm, fan_tot_eff):
    """convert and return inputs for E+ in pascal and m3/s"""
    bhp = watts2bhp(watts)
    return bhp2pascal(bhp, cfm, fan_tot_eff)


def fanpower_bhp(ddtt):
    """return fan power in bhp given the fan IDF object"""
    from eppy.bunch_subclass import (
        BadEPFieldError,
    )  # here to prevent circular dependency

    try:
        fan_tot_eff = ddtt.Fan_Total_Efficiency  # from V+ V8.7.0 onwards
    except BadEPFieldError as e:
        fan_tot_eff = ddtt.Fan_Efficiency
    pascal = float(ddtt.Pressure_Rise)
    if str(ddtt.Maximum_Flow_Rate).lower() == "autosize":
        # str can fail with unicode chars :-(
        return "autosize"
    else:
        m3s = float(ddtt.Maximum_Flow_Rate)
    return fan_bhp(fan_tot_eff, pascal, m3s)


def fanpower_watts(ddtt):
    """return fan power in bhp given the fan IDF object"""
    from eppy.bunch_subclass import (
        BadEPFieldError,
    )  # here to prevent circular dependency

    try:
        fan_tot_eff = ddtt.Fan_Total_Efficiency  # from V+ V8.7.0 onwards
    except BadEPFieldError as e:
        fan_tot_eff = ddtt.Fan_Efficiency
    pascal = float(ddtt.Pressure_Rise)
    if str(ddtt.Maximum_Flow_Rate).lower() == "autosize":
        # str can fail with unicode chars :-(
        return "autosize"
    else:
        m3s = float(ddtt.Maximum_Flow_Rate)
    return fan_watts(fan_tot_eff, pascal, m3s)


def fan_maxcfm(ddtt):
    """return the fan max cfm"""
    if str(ddtt.Maximum_Flow_Rate).lower() == "autosize":
        # str can fail with unicode chars :-(
        return "autosize"
    else:
        m3s = float(ddtt.Maximum_Flow_Rate)
        return m3s2cfm(m3s)



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for some functions of modeleditor"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO
import pytest
from eppy import modeleditor
import eppy.idfreader as idfreader
import eppy.snippet as snippet
from eppy.iddcurrent import iddcurrent

idfsnippet = snippet.idfsnippet
iddsnippet = iddcurrent.iddtxt
idffhandle = StringIO(idfsnippet)
iddfhandle = StringIO(iddsnippet)
bunchdt, data, commdct, idd_index = idfreader.idfreader(idffhandle, iddfhandle, None)


def test_addobject():
    """py.test for addobject"""
    thedata = (
        # key, aname, fielddict
        ("ZONE", None, dict(Name="Gumby", X_Origin=50)),
        ("ZONE", "karamba", {}),  # key, aname, fielddict
        ("ZONE", None, {}),  # key, aname, fielddict
        # key, aname, fielddict
        ("ZONE", None, dict(Name="Gumby", X_Origin=50)),
    )
    for key, aname, fielddict in thedata:
        result = modeleditor.addobject(
            bunchdt, data, commdct, key, None, aname, **fielddict
        )
        assert bunchdt[key][-1].key == key  # wierd, but correct :-)
        if aname:
            assert data.dt[key][-1][1] == aname
            assert bunchdt[key][-1].Name == aname
        if fielddict:
            for kkey, value in fielddict.items():
                assert bunchdt[key][-1][kkey] == value


def test_addobject1():
    """py.test for addobject"""
    thedata = (("ZONE", {"Name": "karamba"}),)  # key, kwargs
    for key, kwargs in thedata:
        result = modeleditor.addobject1(bunchdt, data, commdct, key, **kwargs)
        aname = kwargs["Name"]
        assert data.dt[key][-1][1] == aname
        assert bunchdt[key][-1].Name == aname


def test_getobject():
    """py.test for getobject"""
    thedata = (
        ("ZONE", "PLENUM-1", bunchdt["ZONE"][0]),  # key, name, theobject
        # key, name, theobject
        ("ZONE", "PLENUM-1".lower(), bunchdt["ZONE"][0]),
        ("ZONE", "PLENUM-A", None),  # key, name, theobject
        (
            "ZONEHVAC:EQUIPMENTCONNECTIONS",
            "SPACE1-1",
            bunchdt["ZONEHVAC:EQUIPMENTCONNECTIONS"][0],
        ),  # key, name, theobject
    )
    for key, name, theobject in thedata:
        result = modeleditor.getobject(bunchdt, key, name)
        assert result == theobject


def test___objecthasfields():
    """py.test for __objecthasfields"""
    thedata = (
        ("ZONE", dict(Name="testzone", X_Origin=32), "testzone", True),
        # key, fielddict, aname, istrue
        ("ZONE", dict(Name="testzone", X_Origin=32), "testzone1", False),
        # key, fielddict, aname, istrue
    )
    for key, fielddict, aname, istrue in thedata:
        idfobject = modeleditor.addobject(
            bunchdt, data, commdct, key, None, **fielddict
        )
        idfobject.Name = aname  # modify the name, to check for a False return
        result = modeleditor.__objecthasfields(
            bunchdt, data, commdct, idfobject, **fielddict
        )
        assert result == istrue


def test_getobjects():
    """py.test for getobjects"""
    thedata = (
        ("ZONE", {"Name": "PLENUM-1"}, 7, bunchdt["ZONE"][0:1]),
        # key, fielddict, places, theobjects
        # ('ZONE', {'Name':'PLENUM-1', 'Volume':283.2},7,bunchdt['ZONE'][0:1]),
        # key, fielddict, places, theobjects
        # ('ZONE', {'Y_Origin':0.}, 7, bunchdt['ZONE']),
        # key, fielddict, places, theobjects
    )
    for key, fielddict, places, theobjects in thedata:
        result = modeleditor.getobjects(bunchdt, data, commdct, key, **fielddict)
        assert result == theobjects


def test_is_retaincase():
    """py.test for is_retaincase"""
    thedata = (
        ("BUILDING", "Name", True),  # key, fieldname, case
        ("BUILDING", "Terrain", False),  # key, fieldname, case
    )
    for key, fieldname, case in thedata:
        idfobject = bunchdt[key][0]
        result = modeleditor.is_retaincase(bunchdt, data, commdct, idfobject, fieldname)
        assert result == case


def test_isfieldvalue():
    """py.test for isfieldvalue"""
    thedata = (
        ("BUILDING", 0, "Name", "Building", 7, True),
        # key, objindex1, fieldname, value, places, isequal
        ("BUILDING", 0, "Name", "BuildinG", 7, False),
        # key, objindex1, fieldname, value, places, isequal
        ("BUILDING", 0, "North_Axis", 30, 7, True),
        # key, objindex1, fieldname, value, places, isequal
        ("BUILDING", 0, "North_Axis", "30", 7, True),
        # key, objindex1, fieldname, value, places, isequal
        ("BUILDING", 0, "North_Axis", 30.001, 7, False),
        # key, objindex1, fieldname, value, places, isequal
        ("BUILDING", 0, "North_Axis", 30.001, 2, True),
        # key, objindex1, fieldname, value, places, isequal
        ("ZONE", 0, "Volume", 283.2, 2, True),
        # key, objindex1, fieldname, value, places, isequal
    )
    for key, objindex1, fieldname, value, places, isequal in thedata:
        idfobject = bunchdt[key][objindex1]
        result = modeleditor.isfieldvalue(
            bunchdt, data, commdct, idfobject, fieldname, value, places
        )
        assert result == isequal


def test_equalfield():
    """py.test for equalfield"""
    thedata = (
        ("BUILDING", 0, 1, "Name", 7, True),
        # key, objindex1, objeindex2, fieldname, places, isequal
        ("BUILDING", 0, 2, "Name", 7, False),
        # key, objindex1, objeindex2, fieldname, places, isequal
        ("BUILDING", 0, 1, "Terrain", 7, True),
        # key, objindex1, objeindex2, fieldname, places, isequal
        ("BUILDING", 0, 1, "Terrain", 7, True),
        # key, objindex1, objeindex2, fieldname, places, isequal
        ("BUILDING", 0, 1, "North_Axis", 7, True),
        # key, objindex1, objeindex2, fieldname, places, isequal
        ("BUILDING", 0, 2, "North_Axis", 2, True),
        # key, objindex1, objeindex2, fieldname, places, isequal
        ("BUILDING", 0, 3, "Maximum_Number_of_Warmup_Days", 7, True),
        ("BUILDING", 0, 3, "Minimum_Number_of_Warmup_Days", 7, False),
        # key, objindex1, objeindex2, fieldname, places, isequal
    )
    for key, objindex1, objindex2, fieldname, places, isequal in thedata:
        idfobject1 = bunchdt[key][objindex1]
        idfobject2 = bunchdt[key][objindex2]
        result = modeleditor.equalfield(
            bunchdt, data, commdct, idfobject1, idfobject2, fieldname, places
        )
        assert result == isequal
    (key, objindex1, objeindex2, fieldname, places, isequal) = (
        "BUILDING",
        0,
        1,
        "Name",
        7,
        True,
    )
    idfobject1 = bunchdt[key][objindex1]
    idfobject2 = bunchdt["ZONE"][objindex2]
    with pytest.raises(modeleditor.NotSameObjectError):
        modeleditor.equalfield(
            bunchdt, data, commdct, idfobject1, idfobject2, fieldname, places
        )


def test_removeextensibles():
    """py.test for removeextensibles"""
    thedata = (
        (
            "BuildingSurface:Detailed".upper(),
            "WALL-1PF",
            [
                "BuildingSurface:Detailed",
                "WALL-1PF",
                "WALL",
                "WALL-1",
                "PLENUM-1",
                "Outdoors",
                "",
                "SunExposed",
                "WindExposed",
                "0.50000",
                "4",
            ],
        ),  # key, objname, rawobject
    )
    for key, objname, rawobject in thedata:
        result = modeleditor.removeextensibles(bunchdt, data, commdct, key, objname)
        assert result.obj == rawobject


def test_newrawobject():
    """py.test for newrawobject"""
    thedata = (
        (
            "zone",
            [
                "ZONE",
                "",
                0.0,
                0.0,
                0.0,
                0.0,
                1,
                1,
                "autocalculate",
                "autocalculate",
                "autocalculate",
                "",
                "",
                "Yes",
            ],
        ),  # key, obj
    )
    for key, obj in thedata:
        result = modeleditor.newrawobject(data, commdct, key)
        assert result == obj


def test_obj2bunch():
    """py.test for obj2bunch"""
    thedata = (
        (
            [
                "ZONE",
                "",
                "0",
                "0",
                "0",
                "0",
                "1",
                "1",
                "autocalculate",
                "autocalculate",
                "autocalculate",
                "",
                "",
                "Yes",
            ]
        ),  # obj
    )
    for obj in thedata:
        key_i = data.dtls.index(obj[0].upper())
        abunch = idfreader.makeabunch(commdct, obj, key_i)
        result = modeleditor.obj2bunch(data, commdct, obj)
        assert result.__repr__() == abunch.__repr__()


def test_iddofobject():
    """py.test of iddofobject"""
    thedata = (
        (
            "VERSION",
            [
                {
                    "idfobj": "Version",
                    "group": "Simulation Parameters",
                    "format": ["singleLine"],
                    "unique-object": [""],
                },
                {
                    "default": ["7.0"],
                    "field": ["Version Identifier"],
                    "required-field": [""],
                },
            ],
        ),  # key, itsidd
    )
    for key, itsidd in thedata:
        result = modeleditor.iddofobject(data, commdct, key)
        try:
            result[0].pop("memo")  # memo is new in version 8.0.0
        except KeyError:
            pass
        assert result == itsidd



# Copyright (c) 2018 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================


"""home of easyopen - to easily open an idf file"""

# ideally this should be in idf_helper
# pytest kept failing since the other routines set the IDD file
# pytest works here in a seperate file, like eppy/runner.py

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
from io import IOBase
from io import StringIO
import eppy
import eppy.modeleditor
import eppy.EPlusInterfaceFunctions.parse_idd
import eppy.runner.run_functions


class MissingIDDException(Exception):
    pass


def cleanupversion(ver):
    """massage the version number so it matches the format of install folder"""
    lst = ver.split(".")
    if len(lst) == 1:
        lst.extend(["0", "0"])
    elif len(lst) == 2:
        lst.extend(["0"])
    elif len(lst) > 2:
        lst = lst[:3]
    lst[2] = "0"  # ensure the 3rd number is 0
    cleanver = ".".join(lst)
    return cleanver


def getiddfile(versionid):
    """find the IDD file of the E+ installation"""
    vlist = versionid.split(".")
    if len(vlist) == 1:
        vlist = vlist + ["0", "0"]
    elif len(vlist) == 2:
        vlist = vlist + ["0"]
    ver_str = "-".join(vlist)
    eplus_exe, _ = eppy.runner.run_functions.install_paths(ver_str)
    eplusfolder = os.path.dirname(eplus_exe)
    iddfile = "{}/Energy+.idd".format(eplusfolder)
    return iddfile


def getoldiddfile(versionid):
    """find the IDD file of the E+ installation
    E+ version 7 and earlier have the idd in /EnergyPlus-7-2-0/bin/Energy+.idd"""
    vlist = versionid.split(".")
    if len(vlist) == 1:
        vlist = vlist + ["0", "0"]
    elif len(vlist) == 2:
        vlist = vlist + ["0"]
    ver_str = "-".join(vlist)
    eplus_exe, _ = eppy.runner.run_functions.install_paths(ver_str)
    eplusfolder = os.path.dirname(eplus_exe)
    iddfile = "{}/bin/Energy+.idd".format(eplusfolder)
    return iddfile


def easyopen(fname, idd=None, epw=None):
    """automatically set idd and open idf file. Uses version from idf to set correct idd
    It will work under the following circumstances:

    - the IDF file should have the VERSION object.
    - Needs  the version of EnergyPlus installed that matches the IDF version.
    - Energyplus should be installed in the default location.

    Parameters
    ----------
    fname : str, StringIO or IOBase
        Filepath IDF file,
        File handle of IDF file open to read
        StringIO with IDF contents within
    idd : str, StringIO or IOBase
        This is an optional argument. easyopen will find the IDD without this arg
        Filepath IDD file,
        File handle of IDD file open to read
        StringIO with IDD contents within
    epw : str
        path name to the weather file. This arg is needed to run EneryPlus from eppy.
    """
    if idd:
        eppy.modeleditor.IDF.setiddname(idd)
        idf = eppy.modeleditor.IDF(fname, epw=epw)
        return idf
    # the rest of the code runs if idd=None
    if isinstance(fname, (IOBase, StringIO)):
        fhandle = fname
    else:
        fhandle = io.open(
            fname, "r", encoding="latin-1"
        )  # latin-1 seems to read most things

    # - get the version number from the idf file
    txt = fhandle.read()
    # try:
    #     txt = txt.decode('latin-1') # latin-1 seems to read most things
    # except AttributeError:
    #     pass
    ntxt = eppy.EPlusInterfaceFunctions.parse_idd.nocomment(txt, "!")
    blocks = ntxt.split(";")
    blocks = [block.strip() for block in blocks]
    bblocks = [block.split(",") for block in blocks]
    bblocks1 = [[item.strip() for item in block] for block in bblocks]
    ver_blocks = [block for block in bblocks1 if block[0].upper() == "VERSION"]
    ver_block = ver_blocks[0]
    versionid = ver_block[1]

    # - get the E+ folder based on version number
    iddfile = getiddfile(versionid)
    if os.path.exists(iddfile):
        pass
        # might be an old version of E+
    else:
        iddfile = getoldiddfile(versionid)
    if os.path.exists(iddfile):
        # if True:
        # - set IDD and open IDF.
        eppy.modeleditor.IDF.setiddname(iddfile)
        if isinstance(fname, (IOBase, StringIO)):
            fhandle.seek(0)
            idf = eppy.modeleditor.IDF(fhandle, epw=epw)
        else:
            idf = eppy.modeleditor.IDF(fname, epw=epw)
        return idf

    else:

        # - can't find IDD -> throw an exception
        astr = "input idf file says E+ version {}. easyopen() cannot find the corresponding idd file '{}'"
        astr = astr.format(versionid, iddfile)
        raise MissingIDDException(astr)



# Copyright (c) 2018 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.test for easyopen"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import pytest

import eppy
from eppy import modeleditor
import eppy.easyopen as easyopen
from eppy.pytest_helpers import do_integration_tests
import eppy.idd_helpers as idd_helpers


from six import StringIO
from importlib import reload


def test_cleanupversion():
    """py.test for cleanupversion"""
    data = (
        ("8.8.0", "8.8.0"),  # ver, expected
        ("8.8.1", "8.8.0"),  # ver, expected
        ("8.8", "8.8.0"),  # ver, expected
        ("8", "8.0.0"),  # ver, expected
        ("", ".0.0"),  # ver, expected
    )
    for ver, expected in data:
        result = easyopen.cleanupversion(ver)
        assert result == expected


@pytest.mark.skipif(
    not do_integration_tests(), reason="$EPPY_INTEGRATION env var not set"
)
def test_easyopen_idfopen():
    """py.test for easyopen"""
    ver = idd_helpers.latestidd()
    txt, result = ("  Version,{};".format(ver), "{}".format(ver))
    fhandle1 = StringIO(txt)
    fhandle2 = StringIO(txt)
    reload(eppy)
    reload(modeleditor)
    reload(easyopen)
    idf1, idf2 = easyopen.easyopen(fhandle1), eppy.openidf(fhandle2)
    for idf in [idf1, idf2]:
        versions = idf.idfobjects["version"]
        version = versions[0]
        ver = version.Version_Identifier
        assert result == ver
    # test with epw=weatherfile
    fhandle1 = StringIO(txt)
    fhandle2 = StringIO(txt)
    epwname = "weatherfile.epw"
    idf1, idf2 = (
        easyopen.easyopen(fhandle1, epw=epwname),
        eppy.openidf(fhandle2, epw=epwname),
    )
    for idf in [idf1, idf2]:
        assert idf.epw == epwname


@pytest.mark.skipif(
    not do_integration_tests(), reason="$EPPY_INTEGRATION env var not set"
)
def test_easyopen_withidd():
    """py.test for easyopen"""
    ver = idd_helpers.latestidd()
    iddfile = easyopen.getiddfile(ver)
    txt, result = ("  Version,{};".format(ver), "{}".format(ver))
    fhandle1 = StringIO(txt)
    fhandle2 = StringIO(txt)
    reload(eppy)
    reload(modeleditor)
    reload(easyopen)
    idf1, idf2 = (
        easyopen.easyopen(fhandle1, idd=iddfile),
        eppy.openidf(fhandle2, idd=iddfile),
    )
    for idf in [idf1, idf2]:
        versions = idf.idfobjects["version"]
        version = versions[0]
        ver = version.Version_Identifier
        assert result == ver
    # test with epw=weatherfile
    fhandle1 = StringIO(txt)
    fhandle2 = StringIO(txt)
    epwname = "weatherfile.epw"
    idf1, idf2 = (
        easyopen.easyopen(fhandle1, idd=iddfile, epw=epwname),
        eppy.openidf(fhandle2, idd=iddfile, epw=epwname),
    )
    for idf in [idf1, idf2]:
        assert idf.epw == epwname



# Copyright (c) 2019 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"py.test for functions in __init__"

import pytest

import eppy
from eppy.pytest_helpers import do_integration_tests
import eppy.idd_helpers as idd_helpers


@pytest.mark.skipif(
    not do_integration_tests(), reason="$EPPY_INTEGRATION env var not set"
)
def test_newidf():
    """py.test for newidf"""
    data = ((idd_helpers.latestidd(), idd_helpers.latestidd()),)  # ver, expected
    for ver, expected in data:
        idf = eppy.newidf(ver)
        result = idf.idfobjects["version".upper()][0]
        assert result.Version_Identifier == expected



# Copyright (c) 2019 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""helper functions for idd"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import eppy.runner.run_functions as run_functions


def latestidd():
    """extract the latest idd installed"""
    pth, _ = run_functions.install_paths(
        version="8.8.0"
    )  # works with any value in version
    dirpth = os.path.dirname(pth)
    dirpth = os.path.dirname(dirpth)
    alldirs = os.listdir(dirpth)
    eplusdirs = [dir for dir in alldirs if dir.startswith("EnergyPlus")]
    maxapp = max(eplusdirs)
    ver = folder2ver(maxapp)
    return ver


def folder2ver(folder):
    """get the version number from the E+ install folder"""
    ver = folder.split("EnergyPlus")[-1]
    ver = ver[1:]
    splitapp = ver.split("-")
    ver = ".".join(splitapp)
    return ver



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""pytest for iddgaps.py"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import eppy.iddgaps as iddgaps


def test_cleaniddfield():
    """pytest for cleaniddfield"""
    data = (
        (
            {
                "field": ["Water Supply Storage Tank Name"],
                "Field": ["Water Supply Storage Tank Name"],
                "object-list": ["WaterStorageTankNames"],
                "type": ["object-list"],
            },
            {
                "field": ["Water Supply Storage Tank Name"],
                "object-list": ["WaterStorageTankNames"],
                "type": ["object-list"],
            },
        ),  # field, newfield
    )
    for field, newfield in data:
        result = iddgaps.cleaniddfield(field)
        assert result == newfield



# Copyright (c) 2012 Santosh Philip
# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""Draw all the  loops in the IDF file.

There are two output files saved in the same location as the idf file:
- idf_file_location/idf_filename.dot
- idf_file_location/idf_filename.png

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys

pathnameto_eplusscripting = "../../"
sys.path.append(pathnameto_eplusscripting)

import argparse
import eppy
import eppy.EPlusInterfaceFunctions
from eppy.EPlusInterfaceFunctions import readidf

# import pydot3k as pydot
import pydot

import eppy.loops as loops


pathnameto_eplusscripting = "../../"
sys.path.append(pathnameto_eplusscripting)


def firstisnode(edge):
    if type(edge[0]) == tuple:
        return True
    else:
        return False


def secondisnode(edge):
    if type(edge[1]) == tuple:
        return True
    else:
        return False


def bothnodes(edge):
    if type(edge[0]) == tuple and type(edge[1]) == tuple:
        return True
    else:
        return False


def dropnodes(edges):
    """draw a graph without the nodes"""
    newedges = []
    added = False
    for edge in edges:
        if bothnodes(edge):
            newtup = (edge[0][0], edge[1][0])
            newedges.append(newtup)
            added = True
        elif firstisnode(edge):
            for edge1 in edges:
                if edge[0] == edge1[1]:
                    newtup = (edge1[0], edge[1])
                    try:
                        newedges.index(newtup)
                    except ValueError as e:
                        newedges.append(newtup)
                    added = True
        elif secondisnode(edge):
            for edge1 in edges:
                if edge[1] == edge1[0]:
                    newtup = (edge[0], edge1[1])
                    try:
                        newedges.index(newtup)
                    except ValueError as e:
                        newedges.append(newtup)
                    added = True
        # gets the hanging nodes - nodes with no connection
        if not added:
            if firstisnode(edge):
                newedges.append((edge[0][0], edge[1]))
            if secondisnode(edge):
                newedges.append((edge[0], edge[1][0]))
        added = False
    return newedges


def makeanode(name):
    return pydot.Node(name, shape="plaintext", label=name)


def makeabranch(name):
    return pydot.Node(name, shape="box3d", label=name)


def makeendnode(name):
    return pydot.Node(
        name, shape="doubleoctagon", label=name, style="filled", fillcolor="#e4e4e4"
    )


def istuple(x):
    return type(x) == tuple


def nodetype(anode):
    """return the type of node"""
    try:
        return anode[1]
    except IndexError as e:
        return None


def edges2nodes(edges):
    """gather the nodes from the edges"""
    nodes = []
    for e1, e2 in edges:
        nodes.append(e1)
        nodes.append(e2)
    nodedict = dict([(n, None) for n in nodes])
    justnodes = list(nodedict.keys())
    # justnodes.sort()
    justnodes = sorted(justnodes, key=lambda x: str(x[0]))
    return justnodes


def makediagram(edges):
    """make the diagram with the edges"""
    graph = pydot.Dot(graph_type="digraph")
    nodes = edges2nodes(edges)
    epnodes = [
        (node, makeanode(node[0])) for node in nodes if nodetype(node) == "epnode"
    ]
    endnodes = [
        (node, makeendnode(node[0])) for node in nodes if nodetype(node) == "EndNode"
    ]
    epbr = [(node, makeabranch(node)) for node in nodes if not istuple(node)]
    nodedict = dict(epnodes + epbr + endnodes)
    for value in list(nodedict.values()):
        graph.add_node(value)
    for e1, e2 in edges:
        graph.add_edge(pydot.Edge(nodedict[e1], nodedict[e2]))
    return graph


def transpose2d(mtx):
    """Transpose a 2d matrix
    [
         [1,2,3],
         [4,5,6]
         ]
     becomes
     [
         [1,4],
         [2,5],
         [3,6]
         ]
    """
    return zip(*mtx)


def makebranchcomponents(data, commdct, anode="epnode"):
    """return the edges jointing the components of a branch"""
    alledges = []

    objkey = "BRANCH"
    cnamefield = "Component %s Name"
    inletfield = "Component %s Inlet Node Name"
    outletfield = "Component %s Outlet Node Name"

    numobjects = len(data.dt[objkey])
    cnamefields = loops.repeatingfields(data, commdct, objkey, cnamefield)
    inletfields = loops.repeatingfields(data, commdct, objkey, inletfield)
    outletfields = loops.repeatingfields(data, commdct, objkey, outletfield)

    inlts = loops.extractfields(data, commdct, objkey, [inletfields] * numobjects)
    cmps = loops.extractfields(data, commdct, objkey, [cnamefields] * numobjects)
    otlts = loops.extractfields(data, commdct, objkey, [outletfields] * numobjects)

    zipped = list(zip(inlts, cmps, otlts))
    tzipped = [transpose2d(item) for item in zipped]
    for i in range(len(data.dt[objkey])):
        tt = tzipped[i]
        # branchname = data.dt[objkey][i][1]
        edges = []
        for t0 in tt:
            edges = edges + [((t0[0], anode), t0[1]), (t0[1], (t0[2], anode))]
        alledges = alledges + edges
    return alledges


def makeairplantloop(data, commdct):
    """make the edges for the airloop and the plantloop"""
    anode = "epnode"
    endnode = "EndNode"

    # in plantloop get:
    #     demand inlet, outlet, branchlist
    #     supply inlet, outlet, branchlist
    plantloops = loops.plantloopfields(data, commdct)
    # splitters
    #     inlet
    #     outlet1
    #     outlet2
    splitters = loops.splitterfields(data, commdct)
    #
    # mixer
    #     outlet
    #     inlet1
    #     inlet2

    mixers = loops.mixerfields(data, commdct)
    #
    # supply barnchlist
    #     branch1 -> inlet, outlet
    #     branch2 -> inlet, outlet
    #     branch3 -> inlet, outlet
    #
    # CONNET INLET OUTLETS
    edges = []

    # get all branches
    branchkey = "branch".upper()
    branches = data.dt[branchkey]
    branch_i_o = {}
    for br in branches:
        br_name = br[1]
        in_out = loops.branch_inlet_outlet(data, commdct, br_name)
        branch_i_o[br_name] = dict(list(zip(["inlet", "outlet"], in_out)))
    # for br_name, in_out in branch_i_o.items():
    #     edges.append(((in_out["inlet"], anode), br_name))
    #     edges.append((br_name, (in_out["outlet"], anode)))

    # instead of doing the branch
    # do the content of the branch
    edges = makebranchcomponents(data, commdct)

    # connect splitter to nodes
    for splitter in splitters:
        # splitter_inlet = inletbranch.node
        splittername = splitter[0]
        inletbranchname = splitter[1]
        splitter_inlet = branch_i_o[inletbranchname]["outlet"]
        # edges = splitter_inlet -> splittername
        edges.append(((splitter_inlet, anode), splittername))
        # splitter_outlets = ouletbranches.nodes
        outletbranchnames = [br for br in splitter[2:]]
        splitter_outlets = [branch_i_o[br]["inlet"] for br in outletbranchnames]
        # edges = [splittername -> outlet for outlet in splitter_outlets]
        moreedges = [(splittername, (outlet, anode)) for outlet in splitter_outlets]
        edges = edges + moreedges

    for mixer in mixers:
        # mixer_outlet = outletbranch.node
        mixername = mixer[0]
        outletbranchname = mixer[1]
        mixer_outlet = branch_i_o[outletbranchname]["inlet"]
        # edges = mixername -> mixer_outlet
        edges.append((mixername, (mixer_outlet, anode)))
        # mixer_inlets = inletbranches.nodes
        inletbranchnames = [br for br in mixer[2:]]
        mixer_inlets = [branch_i_o[br]["outlet"] for br in inletbranchnames]
        # edges = [mixername -> inlet for inlet in mixer_inlets]
        moreedges = [((inlet, anode), mixername) for inlet in mixer_inlets]
        edges = edges + moreedges

    # connect demand and supply side
    # for plantloop in plantloops:
    #     supplyinlet = plantloop[1]
    #     supplyoutlet = plantloop[2]
    #     demandinlet = plantloop[4]
    #     demandoutlet = plantloop[5]
    #     # edges = [supplyoutlet -> demandinlet, demandoutlet -> supplyinlet]
    #     moreedges = [((supplyoutlet, endnode), (demandinlet, endnode)),
    #         ((demandoutlet, endnode), (supplyinlet, endnode))]
    #     edges = edges + moreedges
    #
    # -----------air loop stuff----------------------
    # from s_airloop2.py
    # Get the demand and supply nodes from 'airloophvac'
    # in airloophvac get:
    #   get branch, supplyinlet, supplyoutlet, demandinlet, demandoutlet
    objkey = "airloophvac".upper()
    fieldlists = [
        [
            "Branch List Name",
            "Supply Side Inlet Node Name",
            "Demand Side Outlet Node Name",
            "Demand Side Inlet Node Names",
            "Supply Side Outlet Node Names",
        ]
    ] * loops.objectcount(data, objkey)
    airloophvacs = loops.extractfields(data, commdct, objkey, fieldlists)
    # airloophvac = airloophvacs[0]

    # in AirLoopHVAC:ZoneSplitter:
    #   get Name, inlet, all outlets
    objkey = "AirLoopHVAC:ZoneSplitter".upper()
    singlefields = ["Name", "Inlet Node Name"]
    fld = "Outlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    zonesplitters = loops.extractfields(data, commdct, objkey, fieldlists)

    # in AirLoopHVAC:SupplyPlenum:
    #   get Name, Zone Name, Zone Node Name, inlet, all outlets
    objkey = "AirLoopHVAC:SupplyPlenum".upper()
    singlefields = ["Name", "Zone Name", "Zone Node Name", "Inlet Node Name"]
    fld = "Outlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    supplyplenums = loops.extractfields(data, commdct, objkey, fieldlists)

    # in AirLoopHVAC:ZoneMixer:
    #   get Name, outlet, all inlets
    objkey = "AirLoopHVAC:ZoneMixer".upper()
    singlefields = ["Name", "Outlet Node Name"]
    fld = "Inlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    zonemixers = loops.extractfields(data, commdct, objkey, fieldlists)

    # in AirLoopHVAC:ReturnPlenum:
    #   get Name, Zone Name, Zone Node Name, outlet, all inlets
    objkey = "AirLoopHVAC:ReturnPlenum".upper()
    singlefields = ["Name", "Zone Name", "Zone Node Name", "Outlet Node Name"]
    fld = "Inlet %s Node Name"
    repeatfields = loops.repeatingfields(data, commdct, objkey, fld)
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    returnplenums = loops.extractfields(data, commdct, objkey, fieldlists)

    # connect room to each equip in equiplist
    # in ZoneHVAC:EquipmentConnections:
    #   get Name, equiplist, zoneairnode, returnnode
    objkey = "ZoneHVAC:EquipmentConnections".upper()
    singlefields = [
        "Zone Name",
        "Zone Conditioning Equipment List Name",
        "Zone Air Node Name",
        "Zone Return Air Node Name",
    ]
    repeatfields = []
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    equipconnections = loops.extractfields(data, commdct, objkey, fieldlists)
    # in ZoneHVAC:EquipmentList:
    #   get Name, all equiptype, all equipnames
    objkey = "ZoneHVAC:EquipmentList".upper()
    singlefields = ["Name"]
    fieldlist = singlefields
    flds = ["Zone Equipment %s Object Type", "Zone Equipment %s Name"]
    repeatfields = loops.repeatingfields(data, commdct, objkey, flds)
    fieldlist = fieldlist + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    equiplists = loops.extractfields(data, commdct, objkey, fieldlists)
    equiplistdct = dict([(ep[0], ep[1:]) for ep in equiplists])
    for key, equips in list(equiplistdct.items()):
        enames = [equips[i] for i in range(1, len(equips), 2)]
        equiplistdct[key] = enames
    # adistuunit -> room
    # adistuunit <- VAVreheat
    # airinlet -> VAVreheat
    # in ZoneHVAC:AirDistributionUnit:
    #   get Name, equiplist, zoneairnode, returnnode
    objkey = "ZoneHVAC:AirDistributionUnit".upper()
    singlefields = ["Name", "Air Terminal Object Type", "Air Terminal Name"]
    repeatfields = []
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    adistuunits = loops.extractfields(data, commdct, objkey, fieldlists)
    # code only for AirTerminal:SingleDuct:VAV:Reheat
    # get airinletnodes for vavreheats
    # in AirTerminal:SingleDuct:VAV:Reheat:
    #   get Name, airinletnode
    adistuinlets = loops.makeadistu_inlets(data, commdct)
    alladistu_comps = []
    for key in list(adistuinlets.keys()):
        objkey = key.upper()
        singlefields = ["Name"] + adistuinlets[key]
        repeatfields = []
        fieldlist = singlefields + repeatfields
        fieldlists = [fieldlist] * loops.objectcount(data, objkey)
        adistu_components = loops.extractfields(data, commdct, objkey, fieldlists)
        alladistu_comps.append(adistu_components)

    # in AirTerminal:SingleDuct:Uncontrolled:
    #   get Name, airinletnode
    objkey = "AirTerminal:SingleDuct:Uncontrolled".upper()
    singlefields = ["Name", "Zone Supply Air Node Name"]
    repeatfields = []
    fieldlist = singlefields + repeatfields
    fieldlists = [fieldlist] * loops.objectcount(data, objkey)
    uncontrolleds = loops.extractfields(data, commdct, objkey, fieldlists)

    anode = "epnode"
    endnode = "EndNode"

    # edges = []

    # connect demand and supply side
    # for airloophvac in airloophvacs:
    #     supplyinlet = airloophvac[1]
    #     supplyoutlet = airloophvac[4]
    #     demandinlet = airloophvac[3]
    #     demandoutlet = airloophvac[2]
    #     # edges = [supplyoutlet -> demandinlet, demandoutlet -> supplyinlet]
    #     moreedges = [((supplyoutlet, endnode), (demandinlet, endnode)),
    #         ((demandoutlet, endnode), (supplyinlet, endnode))]
    #     edges = edges + moreedges

    # connect zonesplitter to nodes
    for zonesplitter in zonesplitters:
        name = zonesplitter[0]
        inlet = zonesplitter[1]
        outlets = zonesplitter[2:]
        edges.append(((inlet, anode), name))
        for outlet in outlets:
            edges.append((name, (outlet, anode)))

    # connect supplyplenum to nodes
    for supplyplenum in supplyplenums:
        name = supplyplenum[0]
        inlet = supplyplenum[3]
        outlets = supplyplenum[4:]
        edges.append(((inlet, anode), name))
        for outlet in outlets:
            edges.append((name, (outlet, anode)))

    # connect zonemixer to nodes
    for zonemixer in zonemixers:
        name = zonemixer[0]
        outlet = zonemixer[1]
        inlets = zonemixer[2:]
        edges.append((name, (outlet, anode)))
        for inlet in inlets:
            edges.append(((inlet, anode), name))

    # connect returnplenums to nodes
    for returnplenum in returnplenums:
        name = returnplenum[0]
        outlet = returnplenum[3]
        inlets = returnplenum[4:]
        edges.append((name, (outlet, anode)))
        for inlet in inlets:
            edges.append(((inlet, anode), name))

    # connect room to return node
    for equipconnection in equipconnections:
        zonename = equipconnection[0]
        returnnode = equipconnection[-1]
        edges.append((zonename, (returnnode, anode)))

    # connect equips to room
    for equipconnection in equipconnections:
        zonename = equipconnection[0]
        zequiplistname = equipconnection[1]
        for zequip in equiplistdct[zequiplistname]:
            edges.append((zequip, zonename))

    # adistuunit <- adistu_component
    for adistuunit in adistuunits:
        unitname = adistuunit[0]
        compname = adistuunit[2]
        edges.append((compname, unitname))

    # airinlet -> adistu_component
    for adistu_comps in alladistu_comps:
        for adistu_comp in adistu_comps:
            name = adistu_comp[0]
            for airnode in adistu_comp[1:]:
                edges.append(((airnode, anode), name))

    # supplyairnode -> uncontrolled
    for uncontrolled in uncontrolleds:
        name = uncontrolled[0]
        airnode = uncontrolled[1]
        edges.append(((airnode, anode), name))

    # edges = edges + moreedges
    return edges


def getedges(fname, iddfile):
    """return the edges of the idf file fname"""
    data, commdct, _idd_index = readidf.readdatacommdct(fname, iddfile=iddfile)
    edges = makeairplantloop(data, commdct)
    return edges


def replace_colon(s, replacewith="__"):
    """replace the colon with something"""
    return s.replace(":", replacewith)


def clean_edges(arg):
    if isinstance(arg, str):
        return replace_colon(arg)
    try:
        return tuple(clean_edges(x) for x in arg)
    except TypeError:  # catch when for loop fails
        return replace_colon(arg)  # not a sequence so just return repr


def make_and_save_diagram(fname, iddfile):
    g = process_idf(fname, iddfile)
    save_diagram(fname, g)


def process_idf(fname, iddfile):
    data, commdct, _iddindex = readidf.readdatacommdct(fname, iddfile=iddfile)
    print("constructing the loops")
    edges = makeairplantloop(data, commdct)
    print("cleaning edges")
    edges = clean_edges(edges)
    print("making the diagram")

    return makediagram(edges)


def save_diagram(fname, g, silent=False):
    dotname = "%s.dot" % (os.path.splitext(fname)[0])
    pngname = "%s.png" % (os.path.splitext(fname)[0])
    g.write(dotname)
    if not silent:
        print("saved file: %s" % (dotname))
    g.write_png(pngname)
    if not silent:
        print("saved file: %s" % (pngname))


def main():
    parser = argparse.ArgumentParser(
        usage=None, description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )
    # need the formatter to print newline from __doc__
    parser.add_argument(
        "idd",
        type=str,
        action="store",
        help="location of idd file = ./somewhere/eplusv8-0-1.idd",
    )
    parser.add_argument(
        "file",
        type=str,
        action="store",
        help="location of idf file = ./somewhere/f1.idf",
    )
    args = parser.parse_args()
    make_and_save_diagram(args.file, args.idd)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Functions to calculate the thermal properties of constructions and materials.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from itertools import product
import warnings


INSIDE_FILM_R = 0.12
OUTSIDE_FILM_R = 0.03


def rvalue(ddtt):
    """
    R value (m2-K/W) of a construction or material.
    thickness (m) / conductivity (W/m-K)
    """
    object_type = ddtt.obj[0]
    if object_type == "Construction":
        rvalue = INSIDE_FILM_R + OUTSIDE_FILM_R
        layers = ddtt.obj[2:]
        field_idd = ddtt.getfieldidd("Outside_Layer")
        validobjects = field_idd["validobjects"]
        for layer in layers:
            found = False
            for key in validobjects:
                try:
                    rvalue += ddtt.theidf.getobject(key, layer).rvalue
                    found = True
                except AttributeError:
                    pass
            if not found:
                raise AttributeError("%s material not found in IDF" % layer)
    elif object_type == "Material":
        thickness = ddtt.obj[ddtt.objls.index("Thickness")]
        conductivity = ddtt.obj[ddtt.objls.index("Conductivity")]
        rvalue = thickness / conductivity
    elif object_type == "Material:AirGap":
        rvalue = ddtt.obj[ddtt.objls.index("Thermal_Resistance")]
    elif object_type == "Material:InfraredTransparent":
        rvalue = 0
    elif object_type == "Material:NoMass":
        rvalue = ddtt.obj[ddtt.objls.index("Thermal_Resistance")]
    elif object_type == "Material:RoofVegetation":
        warnings.warn(
            "Material:RoofVegetation thermal properties are based on dry soil",
            UserWarning,
        )
        thickness = ddtt.obj[ddtt.objls.index("Thickness")]
        conductivity = ddtt.obj[ddtt.objls.index("Conductivity_of_Dry_Soil")]
        rvalue = thickness / conductivity
    else:
        raise AttributeError("%s rvalue property not implemented" % object_type)
    return rvalue


def ufactor(ddtt):
    """
    U factor (W/m2-K) of a construction or material.
    1 / R value (W/K)
    """
    return 1 / rvalue(ddtt)


def ufactor_ip(ddtt):
    """
    U factor (BTU/(h °F ft^2)) of a construction or material.
    1 / R value (ft^2 °F hr/Btu)
    """
    # quick fix for Santosh. Needs to thought thru
    mult = 0.076 / 0.429  # base on doing conversion in the table report
    return ufactor(ddtt) * mult


def rvalue_ip(ddtt):
    """return R value in IP units"""
    # quick fix for Santosh. Needs to thought thru
    return 1 / ufactor_ip(ddtt)


def heatcapacity(ddtt):
    """
    Heat capacity (kJ/m2-K) of a construction or material.
    thickness (m) * density (kg/m3) * specific heat (J/kg-K) * 0.001
    """
    object_type = ddtt.obj[0]
    if object_type == "Construction":
        heatcapacity = 0
        layers = ddtt.obj[2:]
        field_idd = ddtt.getfieldidd("Outside_Layer")
        validobjects = field_idd["validobjects"]
        for layer in layers:
            found = False
            for key in validobjects:
                try:
                    heatcapacity += ddtt.theidf.getobject(key, layer).heatcapacity
                    found = True
                except AttributeError:
                    pass
            if not found:
                raise AttributeError("%s material not found in IDF" % layer)
    elif object_type == "Material":
        thickness = ddtt.obj[ddtt.objls.index("Thickness")]
        density = ddtt.obj[ddtt.objls.index("Density")]
        specificheat = ddtt.obj[ddtt.objls.index("Specific_Heat")]
        heatcapacity = thickness * density * specificheat * 0.001
    elif object_type == "Material:AirGap":
        heatcapacity = 0
    elif object_type == "Material:InfraredTransparent":
        heatcapacity = 0
    elif object_type == "Material:NoMass":
        warnings.warn(
            "Material:NoMass materials included in heat capacity calculation",
            UserWarning,
        )
        heatcapacity = 0
    elif object_type == "Material:RoofVegetation":
        warnings.warn(
            "Material:RoofVegetation thermal properties are based on dry soil",
            UserWarning,
        )
        thickness = ddtt.obj[ddtt.objls.index("Thickness")]
        density = ddtt.obj[ddtt.objls.index("Density_of_Dry_Soil")]
        specificheat = ddtt.obj[ddtt.objls.index("Specific_Heat_of_Dry_Soil")]
        heatcapacity = thickness * density * specificheat * 0.001
    else:
        raise AttributeError("%s has no heatcapacity property" % object_type)
    return heatcapacity



import os
import shutil

import pytest

from eppy import modeleditor
from eppy.modeleditor import IDF
from eppy.runner.run_functions import paths_from_version, EnergyPlusRunError

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

RESOURCES_DIR = os.path.join(THIS_DIR, os.pardir, "resources")

IDD_FILES = os.path.join(RESOURCES_DIR, "iddfiles")
IDF_FILES = os.path.join(RESOURCES_DIR, "idffiles")


@pytest.mark.xfail
def test_reproduce_run_issue():
    """This is for use as a debugging tool.

    Add the files used in the run as reported/provided by a user.
    Make any additional changes required for reproducing/diagnosing the issue.
    """
    # update the following four lines if necessary
    ep_version = "8-9-0"
    idffile = "V8_9/smallfile.idf"
    iddfile = "Energy+V8_9_0.idd"
    epwfile = "USA_IL_Chicago-OHare.Intl.AP.725300_TMY3.epw"

    _, eplus_home = paths_from_version(ep_version)
    idfname = os.path.join(IDF_FILES, idffile)
    iddfile = os.path.join(IDD_FILES, iddfile)
    epwfile = os.path.join(eplus_home, "WeatherData", epwfile)
    modeleditor.IDF.setiddname(iddfile, testing=True)
    idf = IDF(idfname, epwfile)
    # make any changes to the IDF here
    try:
        # Add any additional `run` kwargs here
        # `ep_version` kwarg is required due to buggy test isolation
        idf.run(output_directory="test_dir", ep_version=ep_version)
        # Add any tests for expected/unexpected outputs here
    except Exception:
        # Add any tests for expected/unexpected exceptions here
        raise
    finally:
        shutil.rmtree("test_dir", ignore_errors=True)



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""images for documentation"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from IPython.display import Image
from IPython.display import (
    display_pretty,
    display_html,
    display_jpeg,
    display_png,
    display_json,
    display_latex,
    display_svg,
)

filemerge = Image(filename="../../docs/images/filemerge.png")
plantloop = Image(filename="../../docs/images/plantloop.png")
idfdiff_path = "../../docs/images/idfdiff.html"
